#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os
import typing
import logging
import itertools
import functools
import fractions
import parsimonious
from parsimonious.nodes import Node as _Node
from . import error
from . import serializable
from . import expression


class DSDLSyntaxError(error.InvalidDefinitionError):
    pass


def parse(text: str, statement_stream_processor: 'StatementStreamProcessor') -> None:
    """
    The entry point of the parser. As the text is being parsed, the parser invokes appropriate
    methods in the statement stream processor.
    """
    pr = _ParseTreeProcessor(statement_stream_processor)
    try:
        pr.parse(text)  # type: ignore

    except error.FrontendError as ex:
        # Inject error location. If this exception is being propagated from a recursive instance, it already has
        # its error location populated, so nothing will happen here.
        ex.set_error_location_if_unknown(line=pr.current_line_number)
        raise ex

    except parsimonious.ParseError as ex:
        raise DSDLSyntaxError('Syntax error', line=int(ex.line())) from None  # type: ignore

    except parsimonious.VisitationError as ex:  # pragma: no cover
        # noinspection PyBroadException
        try:
            line = int(ex.original_class.line())
        except Exception:
            line = pr.current_line_number
        # Treat as internal because all intentional errors are not wrapped into VisitationError.
        assert line > 0
        raise error.InternalError(str(ex), line=line)


class StatementStreamProcessor:
    """
    This interface must be implemented by the logic that sits on top of the transformer.
    The methods are invoked immediately as corresponding statements are encountered within the
    processed DSDL definition.
    This interface can be used to construct a more abstract intermediate representation of the processed text.
    """
    def on_constant(self,
                    constant_type: serializable.SerializableType,
                    name: str,
                    value: expression.Any) -> None:
        raise NotImplementedError  # pragma: no cover

    def on_field(self, field_type: serializable.SerializableType, name: str) -> None:
        raise NotImplementedError  # pragma: no cover

    def on_padding_field(self, padding_field_type: serializable.VoidType) -> None:
        raise NotImplementedError  # pragma: no cover

    def on_directive(self,
                     line_number: int,
                     directive_name: str,
                     associated_expression_value: typing.Optional[expression.Any]) -> None:
        raise NotImplementedError  # pragma: no cover

    def on_service_response_marker(self) -> None:
        """The correctness of the marker placement is not validated by the caller."""
        raise NotImplementedError  # pragma: no cover

    def resolve_top_level_identifier(self, name: str) -> expression.Any:
        """Must throw an appropriate exception if the reference cannot be resolved."""
        raise NotImplementedError  # pragma: no cover

    def resolve_versioned_data_type(self, name: str, version: serializable.Version) -> serializable.CompositeType:
        """Must throw an appropriate exception if the data type is not found."""
        raise NotImplementedError  # pragma: no cover


_logger = logging.getLogger(__name__)


_Children = typing.Tuple[typing.Any, ...]
_VisitorHandler = typing.Callable[['_ParseTreeProcessor', _Node, _Children], typing.Any]
_PrimitiveTypeConstructor = typing.Callable[[serializable.PrimitiveType.CastMode], serializable.PrimitiveType]


def _logged_transformation(fun: _VisitorHandler) -> _VisitorHandler:
    """
    Simply logs the resulting transformation upon its completion.
    """
    @functools.wraps(fun)
    def wrapper(self: '_ParseTreeProcessor', node: _Node, children: _Children) -> typing.Any:
        result = '<TRANSFORMATION FAILED>'  # type: typing.Any
        try:
            result = fun(self, node, children)
            return result
        finally:
            _logger.debug('Transformation: %s(%s) --> %r (source text: %r)',
                          node.expr_name, _print_node(children), result, node.text)

    return wrapper


def _make_typesafe_child_lifter(expected_type: typing.Type[object], logged: bool = False) -> _VisitorHandler:
    def visitor_handler(_self: '_ParseTreeProcessor', _n: _Node, children: _Children) -> typing.Any:
        sole_child, = children
        assert isinstance(sole_child, expected_type), \
            'The child should have been of type %r, not %r: %r' % (expected_type, type(sole_child), sole_child)
        return sole_child

    return _logged_transformation(visitor_handler) if logged else visitor_handler


def _make_binary_operator_handler(operator: expression.BinaryOperator[expression.OperatorOutput]) -> _VisitorHandler:
    return lambda _self, _node, _children: operator


# noinspection PyMethodMayBeStatic
class _ParseTreeProcessor(parsimonious.NodeVisitor):
    """
    This class processes the parse tree, evaluates the expressions and emits a high-level representation
    of the processed description. Essentially it does most of the ground work related to supporting the DSDL
    language, which is bad because it makes the class unnecessarily complex and hard to maintain. Shall it be
    needed to extend the language, please consider refactoring the logic by adding an intermediate abstract
    syntax tree in order to separate the semantic analysis from the grammar-related logic. If that is done,
    expression evaluation will be performed at the AST level rather than at the parse tree level, as it is
    done currently.
    """
    # Populating the default grammar (see the NodeVisitor API).
    grammar = parsimonious.Grammar(  # type: ignore
        open(os.path.join(os.path.dirname(__file__), 'grammar.parsimonious')).read())

    # Intentional exceptions that shall not be treated as parse errors.
    # Beware that those might be propagated from recursive parser instances!
    unwrapped_exceptions = error.FrontendError,  # type: ignore

    def __init__(self, statement_stream_processor: StatementStreamProcessor):
        assert isinstance(statement_stream_processor, StatementStreamProcessor)
        self._statement_stream_processor = statement_stream_processor   # type: StatementStreamProcessor
        self._current_line_number = 1   # Lines are numbered from one

    @property
    def current_line_number(self) -> int:
        assert self._current_line_number > 0
        return self._current_line_number

    def generic_visit(self, node: _Node, children: typing.Sequence[typing.Any]) -> typing.Any:
        """If the node has children, replace the node with them."""
        return tuple(children) or node

    def visit_end_of_line(self, _n: _Node, _c: _Children) -> None:
        self._current_line_number += 1

    # ================================================== Statements ==================================================

    visit_statement           = _make_typesafe_child_lifter(type(None))  # Make sure all sub-nodes have been handled,
    visit_statement_attribute = _make_typesafe_child_lifter(type(None))  # because processing terminates here; these
    visit_statement_directive = _make_typesafe_child_lifter(type(None))  # nodes are above the top level.

    def visit_statement_constant(self, _n: _Node, children: _Children) -> None:
        constant_type, _sp0, name, _sp1, _eq, _sp2, exp = children
        assert isinstance(constant_type, serializable.SerializableType) and isinstance(name, str) and name
        assert isinstance(exp, expression.Any)
        self._statement_stream_processor.on_constant(constant_type, name, exp)

    def visit_statement_field(self, _n: _Node, children: _Children) -> None:
        field_type, _space, name = children
        assert isinstance(field_type, serializable.SerializableType) and isinstance(name, str) and name
        self._statement_stream_processor.on_field(field_type, name)

    def visit_statement_padding_field(self, _n: _Node, children: _Children) -> None:
        void_type = children[0]
        assert isinstance(void_type, serializable.VoidType)
        self._statement_stream_processor.on_padding_field(void_type)

    def visit_statement_service_response_marker(self, _n: _Node, _c: _Children) -> None:
        self._statement_stream_processor.on_service_response_marker()

    def visit_statement_directive_with_expression(self, _n: _Node, children: _Children) -> None:
        _at, name, _space, exp = children
        assert isinstance(name, str) and name and isinstance(exp, expression.Any)
        self._statement_stream_processor.on_directive(line_number=self.current_line_number,
                                                      directive_name=name,
                                                      associated_expression_value=exp)

    def visit_statement_directive_without_expression(self, _n: _Node, children: _Children) -> None:
        _at, name = children
        assert isinstance(name, str) and name
        self._statement_stream_processor.on_directive(line_number=self.current_line_number,
                                                      directive_name=name,
                                                      associated_expression_value=None)

    def visit_identifier(self, node: _Node, _c: _Children) -> str:
        assert isinstance(node.text, str) and node.text
        return node.text

    # ================================================== Data types ==================================================

    visit_type           = _make_typesafe_child_lifter(serializable.SerializableType)
    visit_type_array     = _make_typesafe_child_lifter(serializable.ArrayType, logged=True)
    visit_type_scalar    = _make_typesafe_child_lifter(serializable.SerializableType, logged=True)
    visit_type_primitive = _make_typesafe_child_lifter(serializable.PrimitiveType)

    visit_type_primitive_name = parsimonious.NodeVisitor.lift_child

    def visit_type_array_variable_inclusive(self, _n: _Node, children: _Children) \
            -> serializable.VariableLengthArrayType:
        element_type, _s0, _bl, _s1, _op, _s2, length, _s3, _br = children
        return serializable.VariableLengthArrayType(element_type, _unwrap_array_capacity(length))

    def visit_type_array_variable_exclusive(self, _n: _Node, children: _Children) \
            -> serializable.VariableLengthArrayType:
        element_type, _s0, _bl, _s1, _op, _s2, length, _s3, _br = children
        return serializable.VariableLengthArrayType(element_type, _unwrap_array_capacity(length) - 1)

    def visit_type_array_fixed(self, _n: _Node, children: _Children) -> serializable.FixedLengthArrayType:
        element_type, _s0, _bl, _s1, length, _s2, _br = children
        return serializable.FixedLengthArrayType(element_type, _unwrap_array_capacity(length))

    def visit_type_versioned(self, _n: _Node, children: _Children) -> serializable.CompositeType:
        name, name_tail, _, version = children
        assert isinstance(name, str) and name and isinstance(version, serializable.Version)
        for _, component in name_tail:
            assert isinstance(component, str)
            name += serializable.CompositeType.NAME_COMPONENT_SEPARATOR + component

        return self._statement_stream_processor.resolve_versioned_data_type(name, version)

    def visit_type_version_specifier(self, _n: _Node, children: _Children) -> serializable.Version:
        major, _, minor = children
        assert isinstance(major, expression.Rational) and isinstance(minor, expression.Rational)
        return serializable.Version(major=major.as_native_integer(),
                                    minor=minor.as_native_integer())

    def visit_type_primitive_truncated(self, _n: _Node, children: _Children) -> serializable.PrimitiveType:
        _kw, _sp, cons = children  # type: _Node, _Node, _PrimitiveTypeConstructor
        return cons(serializable.PrimitiveType.CastMode.TRUNCATED)

    def visit_type_primitive_saturated(self, _n: _Node, children: _Children) -> serializable.PrimitiveType:
        _, cons = children  # type: _Node, _PrimitiveTypeConstructor
        return cons(serializable.PrimitiveType.CastMode.SATURATED)

    def visit_type_primitive_name_boolean(self, _n: _Node, _c: _Children) -> _PrimitiveTypeConstructor:
        return lambda cm: serializable.BooleanType(cm)     # lambda is only needed to make mypy shut up

    def visit_type_primitive_name_unsigned_integer(self, _n: _Node, children: _Children) -> _PrimitiveTypeConstructor:
        return lambda cm: serializable.UnsignedIntegerType(children[-1], cm)

    def visit_type_primitive_name_signed_integer(self, _n: _Node, children: _Children) -> _PrimitiveTypeConstructor:
        return lambda cm: serializable.SignedIntegerType(children[-1], cm)

    def visit_type_primitive_name_floating_point(self, _n: _Node, children: _Children) -> _PrimitiveTypeConstructor:
        return lambda cm: serializable.FloatType(children[-1], cm)

    def visit_type_void(self, _n: _Node, children: _Children) -> serializable.VoidType:
        _, width = children
        assert isinstance(width, int)
        return serializable.VoidType(width)

    def visit_type_bit_length_suffix(self, node: _Node, _c: _Children) -> int:
        return int(node.text)

    # ================================================== Expressions ==================================================

    visit_expression = parsimonious.NodeVisitor.lift_child

    visit_op2_log = parsimonious.NodeVisitor.lift_child
    visit_op2_cmp = parsimonious.NodeVisitor.lift_child
    visit_op2_bit = parsimonious.NodeVisitor.lift_child
    visit_op2_add = parsimonious.NodeVisitor.lift_child
    visit_op2_mul = parsimonious.NodeVisitor.lift_child
    visit_op2_exp = parsimonious.NodeVisitor.lift_child

    def visit_expression_list(self, _n: _Node, children: _Children) -> typing.Tuple[expression.Any, ...]:
        out = []    # type: typing.List[expression.Any]
        if children:
            children = children[0]
            assert len(children) == 2
            out = [children[0]]
            for _, _, _, exp in children[1]:
                out.append(exp)

        assert all(map(lambda x: isinstance(x, expression.Any), out))
        return tuple(out)

    @_logged_transformation
    def visit_expression_parenthesized(self, _n: _Node, children: _Children) -> expression.Any:
        _, _, exp, _, _ = children
        assert isinstance(exp, expression.Any)
        return exp

    def visit_expression_atom(self, _n: _Node, children: _Children) -> expression.Any:
        atom, = children
        if isinstance(atom, str):   # Identifier resolution
            new_atom = self._statement_stream_processor.resolve_top_level_identifier(atom)
            if not isinstance(new_atom, expression.Any):
                raise error.InternalError('Identifier %r resolved as %r, expected expression' %
                                          (atom, type(new_atom)))  # pragma: no cover
            _logger.debug('Identifier resolution: %r --> %s', atom, new_atom.TYPE_NAME)
            atom = new_atom
            del new_atom

        assert isinstance(atom, expression.Any)
        return atom

    def _visit_binary_operator_chain(self, _n: _Node, children: _Children) -> expression.Any:
        left = children[0]
        assert isinstance(left, expression.Any)
        for _, operator, _, right in children[1]:
            assert callable(operator)
            left = operator(left, right)
            assert isinstance(left, expression.Any)
        return left

    # Operators are handled through different grammar rules for precedence management purposes.
    # At the time of evaluation there is no point keeping them separate.
    visit_ex_attribute      = _visit_binary_operator_chain
    visit_ex_exponential    = _visit_binary_operator_chain
    visit_ex_multiplicative = _visit_binary_operator_chain
    visit_ex_additive       = _visit_binary_operator_chain
    visit_ex_bitwise        = _visit_binary_operator_chain
    visit_ex_comparison     = _visit_binary_operator_chain
    visit_ex_logical        = _visit_binary_operator_chain

    # These are implemented via unary forms, no handling required.
    visit_ex_logical_not = parsimonious.NodeVisitor.lift_child
    visit_ex_inversion   = parsimonious.NodeVisitor.lift_child

    def visit_op1_form_log_not(self, _n: _Node, children: _Children) -> expression.Any:
        _op, _, exp = children
        assert isinstance(_op, _Node) and isinstance(exp, expression.Any)
        return expression.logical_not(exp)

    def visit_op1_form_inv_pos(self, _n: _Node, children: _Children) -> expression.Any:
        _op, _, exp = children
        assert isinstance(_op, _Node) and isinstance(exp, expression.Any)
        return expression.positive(exp)

    def visit_op1_form_inv_neg(self, _n: _Node, children: _Children) -> expression.Any:
        _op, _, exp = children
        assert isinstance(_op, _Node) and isinstance(exp, expression.Any)
        return expression.negative(exp)

    visit_op2_log_or  = _make_binary_operator_handler(expression.logical_or)
    visit_op2_log_and = _make_binary_operator_handler(expression.logical_and)
    visit_op2_cmp_equ = _make_binary_operator_handler(expression.equal)
    visit_op2_cmp_neq = _make_binary_operator_handler(expression.not_equal)
    visit_op2_cmp_leq = _make_binary_operator_handler(expression.less_or_equal)
    visit_op2_cmp_geq = _make_binary_operator_handler(expression.greater_or_equal)
    visit_op2_cmp_lss = _make_binary_operator_handler(expression.less)
    visit_op2_cmp_grt = _make_binary_operator_handler(expression.greater)
    visit_op2_bit_or  = _make_binary_operator_handler(expression.bitwise_or)
    visit_op2_bit_xor = _make_binary_operator_handler(expression.bitwise_xor)
    visit_op2_bit_and = _make_binary_operator_handler(expression.bitwise_and)
    visit_op2_add_add = _make_binary_operator_handler(expression.add)
    visit_op2_add_sub = _make_binary_operator_handler(expression.subtract)
    visit_op2_mul_mul = _make_binary_operator_handler(expression.multiply)
    visit_op2_mul_div = _make_binary_operator_handler(expression.divide)
    visit_op2_mul_mod = _make_binary_operator_handler(expression.modulo)
    visit_op2_exp_pow = _make_binary_operator_handler(expression.power)

    def visit_op2_attrib(self, _n: _Node, _c: _Children) -> expression.AttributeOperator[expression.Any]:
        return expression.attribute

    # ================================================== Literals ==================================================

    visit_literal         = _make_typesafe_child_lifter(expression.Any, logged=True)
    visit_literal_boolean = _make_typesafe_child_lifter(expression.Boolean)
    visit_literal_string  = _make_typesafe_child_lifter(expression.String)

    def visit_literal_set(self, _n: _Node, children: _Children) -> expression.Set:
        _, _, exp_list, _, _ = children
        assert all(map(lambda x: isinstance(x, expression.Any), exp_list))
        return expression.Set(exp_list)

    def visit_literal_real(self, node: _Node, _c: _Children) -> expression.Rational:
        return expression.Rational(fractions.Fraction(node.text.replace('_', '')))

    def visit_literal_integer(self, node: _Node, _c: _Children) -> expression.Rational:
        return expression.Rational(int(node.text.replace('_', ''), base=0))

    def visit_literal_integer_decimal(self, node: _Node, _c: _Children) -> expression.Rational:
        return expression.Rational(int(node.text.replace('_', '')))

    def visit_literal_boolean_true(self, _n: _Node, _c: _Children) -> expression.Boolean:
        return expression.Boolean(True)

    def visit_literal_boolean_false(self, _n: _Node, _c: _Children) -> expression.Boolean:
        return expression.Boolean(False)

    def visit_literal_string_single_quoted(self, node: _Node, _c: _Children) -> expression.String:
        return _parse_string_literal(node.text)

    def visit_literal_string_double_quoted(self, node: _Node, _c: _Children) -> expression.String:
        return _parse_string_literal(node.text)


#
# Internal helper functions.
#
def _print_node(n: typing.Any) -> str:
    """Simple printing helper; the default printing method from Parsimonious is no good."""
    if isinstance(n, _Node):
        return '%s=%r%s' % (
            n.expr.name or '<anonymous>',
            n.text,
            _print_node(n.children) if n.children else ''
        )
    elif isinstance(n, (list, tuple)):
        return '[%s]' % ', '.join(map(_print_node, n))
    else:
        return repr(n)


def _unwrap_array_capacity(ex: expression.Any) -> int:
    assert isinstance(ex, expression.Any)
    if isinstance(ex, expression.Rational):
        out = ex.as_native_integer()
        assert isinstance(out, int)     # Oh mypy, why are you so weird
        return out
    else:
        raise error.InvalidDefinitionError('Array capacity expression must yield a rational, not %s' %
                                           ex.TYPE_NAME)


def _parse_string_literal(literal: str) -> expression.String:
    assert literal[0] == literal[-1]
    assert literal[0] in '\'\"'
    assert len(literal) >= 2

    quote_symbol = literal[0]
    iterator = iter(literal[1:-1])

    def _next_symbol() -> str:
        try:
            s = next(iterator)  # type: str
        except StopIteration:
            return ''

        if s != '\\':
            assert s != quote_symbol, 'Unescaped quotes cannot appear inside string literals. Bad grammar?'
            return s

        s = next(iterator)
        if s in 'uU':
            h = ''
            for _ in range(4 if s.islower() else 8):
                s = next(iterator).lower()
                if s not in '0123456789abcdef':
                    raise DSDLSyntaxError('Invalid hex character: %r' % s)
                h += s
            return chr(int(h, 16))

        try:
            return {
                'r':  '\r',
                'n':  '\n',
                't':  '\t',
                '"':  '"',
                "'":  "'",
                '\\': '\\',
            }[s.lower()]
        except KeyError:
            raise DSDLSyntaxError('Invalid escape sequence') from None

    out = ''
    for index in itertools.count():  # pragma: no branch
        try:
            symbol = _next_symbol()
        except DSDLSyntaxError as ex:
            raise DSDLSyntaxError('The string literal is malformed after index %d: %s' % (index, ex.text))
        except StopIteration:
            raise DSDLSyntaxError('Unexpected end of string literal after index %d' % index) from None
        else:
            if len(symbol) == 0:
                break
            else:
                assert len(symbol) == 1
                out += symbol

    return expression.String(out)


def _unittest_parse_string_literal() -> None:
    from pytest import raises

    def once(literal: str, value: str) -> None:
        assert _parse_string_literal(literal).native_value == value

    def auto_repr(text: str) -> None:
        r = repr(text)
        for x in range(256):
            r = r.replace(r'\x%02x' % x, r'\u00%02x' % x)
        once(r, text)

    auto_repr('')
    auto_repr('123')
    auto_repr('"')
    auto_repr('"')
    auto_repr('\n')
    auto_repr('\u0000\u0001\U000000ff')

    for a in range(256):
        as_hex = '%02x' % a
        auto_repr('\\u' + as_hex * 2)
        auto_repr('\"\'\\u' + as_hex * 2)
        auto_repr('\\U' + as_hex * 4)

        if chr(a).lower() not in '0123456789abcdef':
            with raises(DSDLSyntaxError, match='.*hex character.*'):
                _parse_string_literal('"\\U0000000%s"' % chr(a))

            with raises(DSDLSyntaxError, match='.*hex character.*'):
                _parse_string_literal("'\\u00%s0'" % chr(a))
        else:
            with raises(DSDLSyntaxError, match='.*expected.*'):
                _parse_string_literal("'\\u%s'" % chr(a))

    with raises(DSDLSyntaxError, match='.*expected.*'):
        _parse_string_literal("'\\u'")

    with raises(DSDLSyntaxError, match='.*expected.*'):
        _parse_string_literal("'\\'")

    with raises(DSDLSyntaxError, match='.*escape.*'):
        _parse_string_literal("'\\z'")

    once('"evening"', 'evening')    # okay we support English, cool
    once('"вечер"', 'вечер')        # and Russian too
    once('"õhtust"', 'õhtust')      # heck, even Estonian



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os
import sys

if sys.version_info[:2] < (3, 5):   # pragma: no cover
    print('A newer version of Python is required', file=sys.stderr)
    sys.exit(1)

__version__ = 0, 5, 0
__license__ = 'MIT'

# Our unorthodox approach to dependency management requires us to apply certain workarounds.
# Here, the objective is to allow our library to import stuff from its third-party dependency directory,
# but at the same time we don't want to interfere with the application that depends on this library.
# So we modify the import lookup path temporarily while the package initialization is in progress;
# when done, we restore the path back to its original value. One implication is that it won't be possible
# to import stuff dynamically after the initialization is finished (e.g., function-local imports won't be
# able to reach the third-party stuff), but we don't care.
_original_sys_path = sys.path
sys.path = [os.path.join(os.path.dirname(__file__), 'third_party')] + sys.path

# Never import anything that is not available here - API stability guarantees are only provided for the exposed items.
from .namespace import read_namespace
from .namespace import PrintOutputHandler

# Error model.
from .error import FrontendError, InvalidDefinitionError, InternalError

# Data type model - meta types.
from .serializable import SerializableType
from .serializable import PrimitiveType
from .serializable import BooleanType
from .serializable import ArithmeticType, IntegerType, SignedIntegerType, UnsignedIntegerType, FloatType
from .serializable import VoidType
from .serializable import ArrayType, FixedLengthArrayType, VariableLengthArrayType
from .serializable import CompositeType, UnionType, StructureType, ServiceType

# Data type model - attributes.
from .serializable import Attribute, Field, PaddingField, Constant

# Data type model - auxiliary.
from .serializable import BitLengthRange, ValueRange, Version

# Expression model.
from .expression import Any
from .expression import Primitive, Boolean, Rational, String
from .expression import Container, Set

sys.path = _original_sys_path



"""Subexpressions that make up a parsed grammar

These do the parsing.

"""
# TODO: Make sure all symbol refs are local--not class lookups or
# anything--for speed. And kill all the dots.

from inspect import getargspec
import re

from six import integer_types, python_2_unicode_compatible
from six.moves import range

from parsimonious.exceptions import ParseError, IncompleteParseError
from parsimonious.nodes import Node, RegexNode
from parsimonious.utils import StrAndRepr

MARKER = object()


def expression(callable, rule_name, grammar):
    """Turn a plain callable into an Expression.

    The callable can be of this simple form::

        def foo(text, pos):
            '''If this custom expression matches starting at text[pos], return
            the index where it stops matching. Otherwise, return None.'''
            if the expression matched:
                return end_pos

    If there child nodes to return, return a tuple::

        return end_pos, children

    If the expression doesn't match at the given ``pos`` at all... ::

        return None

    If your callable needs to make sub-calls to other rules in the grammar or
    do error reporting, it can take this form, gaining additional arguments::

        def foo(text, pos, cache, error, grammar):
            # Call out to other rules:
            node = grammar['another_rule'].match_core(text, pos, cache, error)
            ...
            # Return values as above.

    The return value of the callable, if an int or a tuple, will be
    automatically transmuted into a :class:`~parsimonious.Node`. If it returns
    a Node-like class directly, it will be passed through unchanged.

    :arg rule_name: The rule name to attach to the resulting
        :class:`~parsimonious.Expression`
    :arg grammar: The :class:`~parsimonious.Grammar` this expression will be a
        part of, to make delegating to other rules possible

    """
    num_args = len(getargspec(callable).args)
    if num_args == 2:
        is_simple = True
    elif num_args == 5:
        is_simple = False
    else:
        raise RuntimeError("Custom rule functions must take either 2 or 5 "
                           "arguments, not %s." % num_args)

    class AdHocExpression(Expression):
        def _uncached_match(self, text, pos, cache, error):
            result = (callable(text, pos) if is_simple else
                      callable(text, pos, cache, error, grammar))

            if isinstance(result, integer_types):
                end, children = result, None
            elif isinstance(result, tuple):
                end, children = result
            else:
                # Node or None
                return result
            return Node(self, text, pos, end, children=children)

        def _as_rhs(self):
            return '{custom function "%s"}' % callable.__name__

    return AdHocExpression(name=rule_name)


@python_2_unicode_compatible
class Expression(StrAndRepr):
    """A thing that can be matched against a piece of text"""

    # Slots are about twice as fast as __dict__-based attributes:
    # http://stackoverflow.com/questions/1336791/dictionary-vs-object-which-is-more-efficient-and-why

    # Top-level expressions--rules--have names. Subexpressions are named ''.
    __slots__ = ['name', 'identity_tuple']

    def __init__(self, name=''):
        self.name = name
        self.identity_tuple = (self.name, )

    def __hash__(self):
        return hash(self.identity_tuple)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.identity_tuple == other.identity_tuple

    def __ne__(self, other):
        return not (self == other)

    def parse(self, text, pos=0):
        """Return a parse tree of ``text``.

        Raise ``ParseError`` if the expression wasn't satisfied. Raise
        ``IncompleteParseError`` if the expression was satisfied but didn't
        consume the full string.

        """
        node = self.match(text, pos=pos)
        if node.end < len(text):
            raise IncompleteParseError(text, node.end, self)
        return node

    def match(self, text, pos=0):
        """Return the parse tree matching this expression at the given
        position, not necessarily extending all the way to the end of ``text``.

        Raise ``ParseError`` if there is no match there.

        :arg pos: The index at which to start matching

        """
        error = ParseError(text)
        node = self.match_core(text, pos, {}, error)
        if node is None:
            raise error
        return node

    def match_core(self, text, pos, cache, error):
        """Internal guts of ``match()``

        This is appropriate to call only from custom rules or Expression
        subclasses.

        :arg cache: The packrat cache::

            {(oid, pos): Node tree matched by object `oid` at index `pos` ...}

        :arg error: A ParseError instance with ``text`` already filled in but
            otherwise blank. We update the error reporting info on this object
            as we go. (Sticking references on an existing instance is faster
            than allocating a new one for each expression that fails.) We
            return None rather than raising and catching ParseErrors because
            catching is slow.

        """
        # TODO: Optimize. Probably a hot spot.
        #
        # Is there a way of looking up cached stuff that's faster than hashing
        # this id-pos pair?
        #
        # If this is slow, think about the array module. It might (or might
        # not!) use more RAM, but it'll likely be faster than hashing things
        # all the time. Also, can we move all the allocs up front?
        #
        # To save space, we have lots of choices: (0) Quit caching whole Node
        # objects. Cache just what you need to reconstitute them. (1) Cache
        # only the results of entire rules, not subexpressions (probably a
        # horrible idea for rules that need to backtrack internally a lot). (2)
        # Age stuff out of the cache somehow. LRU? (3) Cuts.
        expr_id = id(self)
        node = cache.get((expr_id, pos), MARKER)  # TODO: Change to setdefault to prevent infinite recursion in left-recursive rules.
        if node is MARKER:
            node = cache[(expr_id, pos)] = self._uncached_match(text,
                                                                pos,
                                                                cache,
                                                                error)

        # Record progress for error reporting:
        if node is None and pos >= error.pos and (
                self.name or getattr(error.expr, 'name', None) is None):
            # Don't bother reporting on unnamed expressions (unless that's all
            # we've seen so far), as they're hard to track down for a human.
            # Perhaps we could include the unnamed subexpressions later as
            # auxiliary info.
            error.expr = self
            error.pos = pos

        return node

    def __str__(self):
        return u'<%s %s>' % (
            self.__class__.__name__,
            self.as_rule())

    def as_rule(self):
        """Return the left- and right-hand sides of a rule that represents me.

        Return unicode. If I have no ``name``, omit the left-hand side.

        """
        rhs = self._as_rhs().strip()
        if rhs.startswith('(') and rhs.endswith(')'):
            rhs = rhs[1:-1]

        return (u'%s = %s' % (self.name, rhs)) if self.name else rhs

    def _unicode_members(self):
        """Return an iterable of my unicode-represented children, stopping
        descent when we hit a named node so the returned value resembles the
        input rule."""
        return [(m.name or m._as_rhs()) for m in self.members]

    def _as_rhs(self):
        """Return the right-hand side of a rule that represents me.

        Implemented by subclasses.

        """
        raise NotImplementedError


class Literal(Expression):
    """A string literal

    Use these if you can; they're the fastest.

    """
    __slots__ = ['literal']

    def __init__(self, literal, name=''):
        super(Literal, self).__init__(name)
        self.literal = literal
        self.identity_tuple = (name, literal)

    def _uncached_match(self, text, pos, cache, error):
        if text.startswith(self.literal, pos):
            return Node(self, text, pos, pos + len(self.literal))

    def _as_rhs(self):
        # TODO: Get backslash escaping right.
        return '"%s"' % self.literal


class TokenMatcher(Literal):
    """An expression matching a single token of a given type

    This is for use only with TokenGrammars.

    """
    def _uncached_match(self, token_list, pos, cache, error):
        if token_list[pos].type == self.literal:
            return Node(self, token_list, pos, pos + 1)


class Regex(Expression):
    """An expression that matches what a regex does.

    Use these as much as you can and jam as much into each one as you can;
    they're fast.

    """
    __slots__ = ['re']

    def __init__(self, pattern, name='', ignore_case=False, locale=False,
                 multiline=False, dot_all=False, unicode=False, verbose=False):
        super(Regex, self).__init__(name)
        self.re = re.compile(pattern, (ignore_case and re.I) |
                                      (locale and re.L) |
                                      (multiline and re.M) |
                                      (dot_all and re.S) |
                                      (unicode and re.U) |
                                      (verbose and re.X))
        self.identity_tuple = (self.name, self.re)

    def _uncached_match(self, text, pos, cache, error):
        """Return length of match, ``None`` if no match."""
        m = self.re.match(text, pos)
        if m is not None:
            span = m.span()
            node = RegexNode(self, text, pos, pos + span[1] - span[0])
            node.match = m  # TODO: A terrible idea for cache size?
            return node

    def _regex_flags_from_bits(self, bits):
        """Return the textual equivalent of numerically encoded regex flags."""
        flags = 'ilmsux'
        return ''.join(flags[i - 1] if (1 << i) & bits else '' for i in range(1, len(flags) + 1))

    def _as_rhs(self):
        # TODO: Get backslash escaping right.
        return '~"%s"%s' % (self.re.pattern,
                            self._regex_flags_from_bits(self.re.flags))


class Compound(Expression):
    """An abstract expression which contains other expressions"""

    __slots__ = ['members']

    def __init__(self, *members, **kwargs):
        """``members`` is a sequence of expressions."""
        super(Compound, self).__init__(kwargs.get('name', ''))
        self.members = members

    def __hash__(self):
        # Note we leave members out of the hash computation, since compounds can get added to
        # sets, then have their members mutated. See RuleVisitor._resolve_refs.
        # Equality should still work, but we want the rules to go into the correct hash bucket.
        return hash((self.__class__, self.name))

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            self.name == other.name and
            self.members == other.members)


class Sequence(Compound):
    """A series of expressions that must match contiguous, ordered pieces of
    the text

    In other words, it's a concatenation operator: each piece has to match, one
    after another.

    """
    def _uncached_match(self, text, pos, cache, error):
        new_pos = pos
        length_of_sequence = 0
        children = []
        for m in self.members:
            node = m.match_core(text, new_pos, cache, error)
            if node is None:
                return None
            children.append(node)
            length = node.end - node.start
            new_pos += length
            length_of_sequence += length
        # Hooray! We got through all the members!
        return Node(self, text, pos, pos + length_of_sequence, children)

    def _as_rhs(self):
        return u'({0})'.format(u' '.join(self._unicode_members()))


class OneOf(Compound):
    """A series of expressions, one of which must match

    Expressions are tested in order from first to last. The first to succeed
    wins.

    """
    def _uncached_match(self, text, pos, cache, error):
        for m in self.members:
            node = m.match_core(text, pos, cache, error)
            if node is not None:
                # Wrap the succeeding child in a node representing the OneOf:
                return Node(self, text, pos, node.end, children=[node])

    def _as_rhs(self):
        return u'({0})'.format(u' / '.join(self._unicode_members()))


class Lookahead(Compound):
    """An expression which consumes nothing, even if its contained expression
    succeeds"""

    # TODO: Merge this and Not for better cache hit ratios and less code.
    # Downside: pretty-printed grammars might be spelled differently than what
    # went in. That doesn't bother me.

    def _uncached_match(self, text, pos, cache, error):
        node = self.members[0].match_core(text, pos, cache, error)
        if node is not None:
            return Node(self, text, pos, pos)

    def _as_rhs(self):
        return u'&%s' % self._unicode_members()[0]


class Not(Compound):
    """An expression that succeeds only if the expression within it doesn't

    In any case, it never consumes any characters; it's a negative lookahead.

    """
    def _uncached_match(self, text, pos, cache, error):
        # FWIW, the implementation in Parsing Techniques in Figure 15.29 does
        # not bother to cache NOTs directly.
        node = self.members[0].match_core(text, pos, cache, error)
        if node is None:
            return Node(self, text, pos, pos)

    def _as_rhs(self):
        # TODO: Make sure this parenthesizes the member properly if it's an OR
        # or AND.
        return u'!%s' % self._unicode_members()[0]


# Quantifiers. None of these is strictly necessary, but they're darn handy.

class Optional(Compound):
    """An expression that succeeds whether or not the contained one does

    If the contained expression succeeds, it goes ahead and consumes what it
    consumes. Otherwise, it consumes nothing.

    """
    def _uncached_match(self, text, pos, cache, error):
        node = self.members[0].match_core(text, pos, cache, error)
        return (Node(self, text, pos, pos) if node is None else
                Node(self, text, pos, node.end, children=[node]))

    def _as_rhs(self):
        return u'%s?' % self._unicode_members()[0]


# TODO: Merge with OneOrMore.
class ZeroOrMore(Compound):
    """An expression wrapper like the * quantifier in regexes."""

    def _uncached_match(self, text, pos, cache, error):
        new_pos = pos
        children = []
        while True:
            node = self.members[0].match_core(text, new_pos, cache, error)
            if node is None or not (node.end - node.start):
                # Node was None or 0 length. 0 would otherwise loop infinitely.
                return Node(self, text, pos, new_pos, children)
            children.append(node)
            new_pos += node.end - node.start

    def _as_rhs(self):
        return u'%s*' % self._unicode_members()[0]


class OneOrMore(Compound):
    """An expression wrapper like the + quantifier in regexes.

    You can also pass in an alternate minimum to make this behave like "2 or
    more", "3 or more", etc.

    """
    __slots__ = ['min']

    # TODO: Add max. It should probably succeed if there are more than the max
    # --just not consume them.

    def __init__(self, member, name='', min=1):
        super(OneOrMore, self).__init__(member, name=name)
        self.min = min

    def _uncached_match(self, text, pos, cache, error):
        new_pos = pos
        children = []
        while True:
            node = self.members[0].match_core(text, new_pos, cache, error)
            if node is None:
                break
            children.append(node)
            length = node.end - node.start
            if length == 0:  # Don't loop infinitely.
                break
            new_pos += length
        if len(children) >= self.min:
            return Node(self, text, pos, new_pos, children)

    def _as_rhs(self):
        return u'%s+' % self._unicode_members()[0]



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import typing
import urllib.parse


class FrontendError(Exception):       # PEP8 says that the "Exception" suffix is redundant and should not be used.
    """
    This exception is raised in case if the front end discovers an error in the DSDL code, or if it encounters
    an internal failure that can be associated with a particular construct in the DSDL definition.
    This is the base class for all exceptions raised by the front-end's inner logic,
    excepting its entry function which also can raise ValueError when provided with incorrect inputs.
    Fields:
        path    Source file path where the error has occurred. Optional, will be None if unknown.
        line    Source file line number where the error has occurred. Optional, will be None if unknown.
                The path is always known if the line number is set.
    """

    def __init__(self,
                 text: str,
                 path: typing.Optional[str] = None,
                 line: typing.Optional[int] = None):
        Exception.__init__(self, text)
        self._path = path
        self._line = line

    def set_error_location_if_unknown(self,
                                      path: typing.Optional[str] = None,
                                      line: typing.Optional[int] = None) -> None:
        """
        Entries that are already known will be left unchanged. This is useful when propagating exceptions through
        recursive instances, e.g., when processing nested definitions.
        """
        if not self._path and path:
            self._path = path

        if not self._line and line:
            self._line = line

    @property
    def path(self) -> typing.Optional[str]:
        return self._path

    @property
    def line(self) -> typing.Optional[int]:
        return self._line

    @property
    def text(self) -> str:
        return Exception.__str__(self)

    def __str__(self) -> str:
        """Returns a nicely formatted error string in a GCC-like format (can be parsed by e.g. Eclipse error parser)"""
        if self.path and self.line:
            return '%s:%d: %s' % (self.path, self.line, self.text)

        if self.path:
            return '%s: %s' % (self.path, self.text)

        return self.text

    def __repr__(self) -> str:
        return self.__class__.__name__ + ': ' + repr(self.__str__())


class InternalError(FrontendError):
    """
    This exception is used to report internal errors in the front end itself that prevented it from
    processing the definitions.
    """
    def __init__(self,
                 text: typing.Optional[str] = None,
                 path: typing.Optional[str] = None,
                 line: typing.Optional[int] = None,
                 culprit: typing.Optional[Exception] = None):
        if culprit is not None:
            report_text = 'PLEASE REPORT AT https://github.com/UAVCAN/pydsdl/issues/new?title=' + \
                          urllib.parse.quote(repr(culprit))
            if text:
                text = text + ' ' + report_text
            else:   # pragma: no cover
                text = report_text

        if not text:
            text = ''

        super(InternalError, self).__init__(text=text, path=path, line=line)


class InvalidDefinitionError(FrontendError):
    """
    This exception is used to point out mistakes and errors in DSDL definitions.
    """
    pass


def _unittest_error() -> None:
    try:
        raise FrontendError('Hello world!')
    except Exception as ex:
        assert str(ex) == 'Hello world!'
        assert repr(ex) == "FrontendError: 'Hello world!'"

    try:
        raise FrontendError('Hello world!', path='path/to/file.uavcan', line=123)
    except Exception as ex:
        assert str(ex) == 'path/to/file.uavcan:123: Hello world!'
        assert repr(ex) == "FrontendError: 'path/to/file.uavcan:123: Hello world!'"

    try:
        raise FrontendError('Hello world!', path='path/to/file.uavcan')
    except Exception as ex:
        assert str(ex) == 'path/to/file.uavcan: Hello world!'
        assert repr(ex) == "FrontendError: 'path/to/file.uavcan: Hello world!'"


def _unittest_internal_error_github_reporting() -> None:
    try:
        raise InternalError(path='FILE_PATH',
                            line=42)
    except FrontendError as ex:
        assert ex.path == 'FILE_PATH'
        assert ex.line == 42
        assert str(ex) == 'FILE_PATH:42: '

    try:
        try:
            try:    # TRY HARDER
                raise InternalError(text='BASE TEXT',
                                    culprit=Exception('ERROR TEXT'))
            except FrontendError as ex:
                ex.set_error_location_if_unknown(path='FILE_PATH')
                raise
        except FrontendError as ex:
            ex.set_error_location_if_unknown(line=42)
            raise
    except FrontendError as ex:
        print(ex)
        assert ex.path == 'FILE_PATH'
        assert ex.line == 42
        # We have to ignore the last couple of characters because Python before 3.7 reprs Exceptions like this:
        #   Exception('ERROR TEXT',)
        # But newer Pythons do it like this:
        #   Exception('ERROR TEXT')
        assert str(ex).startswith(
            'FILE_PATH:42: BASE TEXT '
            'PLEASE REPORT AT https://github.com/UAVCAN/pydsdl/issues/new?title=Exception%28%27ERROR%20TEXT%27'
        )

    try:
        raise InternalError(text='BASE TEXT',
                            path='FILE_PATH')
    except FrontendError as ex:
        assert str(ex) == 'FILE_PATH: BASE TEXT'



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import typing
import logging
from . import serializable
from . import expression
from . import error
from . import dsdl_definition
from . import parser
from . import data_schema_builder
from . import port_id_ranges


class AssertionCheckFailureError(error.InvalidDefinitionError):
    pass


class UndefinedDataTypeError(error.InvalidDefinitionError):
    pass


class UndefinedIdentifierError(error.InvalidDefinitionError):
    pass


class InvalidDirectiveError(error.InvalidDefinitionError):
    pass


_logger = logging.getLogger(__name__)


class DataTypeBuilder(parser.StatementStreamProcessor):
    def __init__(self,
                 definition:                      dsdl_definition.DSDLDefinition,
                 lookup_definitions:              typing.Iterable[dsdl_definition.DSDLDefinition],
                 print_output_handler:            typing.Callable[[int, str], None],
                 allow_unregulated_fixed_port_id: bool):
        self._definition = definition
        self._lookup_definitions = list(lookup_definitions)
        self._print_output_handler = print_output_handler
        self._allow_unregulated_fixed_port_id = allow_unregulated_fixed_port_id

        assert isinstance(self._definition, dsdl_definition.DSDLDefinition)
        assert all(map(lambda x: isinstance(x, dsdl_definition.DSDLDefinition), lookup_definitions))
        assert callable(self._print_output_handler)
        assert isinstance(self._allow_unregulated_fixed_port_id, bool)

        self._structs = [data_schema_builder.DataSchemaBuilder()]
        self._is_deprecated = False

    def finalize(self) -> serializable.CompositeType:
        if len(self._structs) == 1:     # Message type
            struct, = self._structs     # type: data_schema_builder.DataSchemaBuilder,
            if struct.union:
                out = serializable.UnionType(
                    name=self._definition.full_name,
                    version=self._definition.version,
                    attributes=struct.attributes,
                    deprecated=self._is_deprecated,
                    fixed_port_id=self._definition.fixed_port_id,
                    source_file_path=self._definition.file_path)  # type: serializable.CompositeType
            else:
                out = serializable.StructureType(
                    name=self._definition.full_name,
                    version=self._definition.version,
                    attributes=struct.attributes,
                    deprecated=self._is_deprecated,
                    fixed_port_id=self._definition.fixed_port_id,
                    source_file_path=self._definition.file_path)
        else:  # Service type
            request, response = self._structs
            # noinspection SpellCheckingInspection
            out = serializable.ServiceType(
                name=self._definition.full_name,            # pozabito vse na svete
                version=self._definition.version,           # serdce zamerlo v grudi
                request_attributes=request.attributes,      # tolko nebo tolko veter
                response_attributes=response.attributes,    # tolko radost vperedi
                request_is_union=request.union,             # tolko nebo tolko veter
                response_is_union=response.union,           # tolko radost vperedi
                deprecated=self._is_deprecated,
                fixed_port_id=self._definition.fixed_port_id,
                source_file_path=self._definition.file_path)

        if not self._allow_unregulated_fixed_port_id:
            port_id = out.fixed_port_id
            if port_id is not None:
                is_service_type = isinstance(out, serializable.ServiceType)
                f = port_id_ranges.is_valid_regulated_service_id if is_service_type else \
                    port_id_ranges.is_valid_regulated_subject_id
                if not f(port_id, out.root_namespace):
                    raise serializable.InvalidFixedPortIDError(
                        'Regulated port ID %r for %s type %r is not valid. '
                        'Consider using allow_unregulated_fixed_port_id.' %
                        (port_id, 'service' if is_service_type else 'message', out.full_name))

        assert isinstance(out, serializable.CompositeType)
        return out

    def on_constant(self,
                    constant_type: serializable.SerializableType,
                    name: str,
                    value: expression.Any) -> None:
        self._structs[-1].add_constant(serializable.Constant(constant_type, name, value))

    def on_field(self, field_type: serializable.SerializableType, name: str) -> None:
        self._structs[-1].add_field(serializable.Field(field_type, name))

    def on_padding_field(self, padding_field_type: serializable.VoidType) -> None:
        self._structs[-1].add_field(serializable.PaddingField(padding_field_type))

    def on_directive(self,
                     line_number: int,
                     directive_name: str,
                     associated_expression_value: typing.Optional[expression.Any]) -> None:
        try:
            handler = {
                'print':      self._on_print_directive,
                'assert':     self._on_assert_directive,
                'union':      self._on_union_directive,
                'deprecated': self._on_deprecated_directive,
            }[directive_name]
        except KeyError:
            raise InvalidDirectiveError('Unknown directive %r' % directive_name)
        else:
            assert callable(handler)
            return handler(line_number, associated_expression_value)

    def on_service_response_marker(self) -> None:
        if len(self._structs) > 1:
            raise error.InvalidDefinitionError('Duplicated service response marker')

        self._structs.append(data_schema_builder.DataSchemaBuilder())
        assert len(self._structs) == 2

    def resolve_top_level_identifier(self, name: str) -> expression.Any:
        # Look only in the current data structure. The lookup cannot cross the service request/response boundary.
        for c in self._structs[-1].constants:
            if c.name == name:
                return c.value

        if name == '_offset_':
            blv = self._structs[-1].compute_bit_length_values()
            assert isinstance(blv, set) and len(blv) > 0 and all(map(lambda x: isinstance(x, int), blv))
            return expression.Set(map(expression.Rational, blv))
        else:
            raise UndefinedIdentifierError('Undefined identifier: %r' % name)

    def resolve_versioned_data_type(self, name: str, version: serializable.Version) -> serializable.CompositeType:
        if serializable.CompositeType.NAME_COMPONENT_SEPARATOR in name:
            full_name = name
        else:
            full_name = serializable.CompositeType.NAME_COMPONENT_SEPARATOR.join([self._definition.full_namespace,
                                                                                  name])
            _logger.info('The full name of a relatively referred type %r reconstructed as %r', name, full_name)

        del name
        found = list(filter(lambda d: d.full_name == full_name and d.version == version, self._lookup_definitions))
        if not found:
            raise UndefinedDataTypeError('Data type %r version %d.%d could not be found' %
                                         (full_name, version.major, version.minor))
        if len(found) > 1:  # pragma: no cover
            raise error.InternalError('Conflicting definitions: %r' % found)

        target_definition = found[0]
        assert isinstance(target_definition, dsdl_definition.DSDLDefinition)
        assert target_definition.full_name == full_name
        assert target_definition.version == version
        # Recursion is cool.
        return target_definition.read(lookup_definitions=self._lookup_definitions,
                                      print_output_handler=self._print_output_handler,
                                      allow_unregulated_fixed_port_id=self._allow_unregulated_fixed_port_id)

    def _on_print_directive(self, line_number: int, value: typing.Optional[expression.Any]) -> None:
        _logger.info('Print directive at %s:%d%s', self._definition.file_path, line_number,
                     (': %s' % value) if value is not None else ' (no value to print)')
        self._print_output_handler(line_number, str(value if value is not None else ''))

    def _on_assert_directive(self, line_number: int, value: typing.Optional[expression.Any]) -> None:
        if isinstance(value, expression.Boolean):
            if not value.native_value:
                raise AssertionCheckFailureError('Assertion check has failed',
                                                 path=self._definition.file_path,
                                                 line=line_number)
            else:
                _logger.debug('Assertion check successful at %s:%d', self._definition.file_path, line_number)
        elif value is None:
            raise InvalidDirectiveError('Assert directive requires an expression')
        else:
            raise InvalidDirectiveError('The assertion check expression must yield a boolean, not %s' %
                                        value.TYPE_NAME)

    def _on_union_directive(self, _ln: int, value: typing.Optional[expression.Any]) -> None:
        if value is not None:
            raise InvalidDirectiveError('The union directive does not expect an expression')

        if self._structs[-1].union:
            raise InvalidDirectiveError('Duplicated union directive')

        if not self._structs[-1].empty:
            raise InvalidDirectiveError('The union directive must be placed before the first '
                                        'attribute definition')

        self._structs[-1].make_union()

    def _on_deprecated_directive(self, _ln: int, value: typing.Optional[expression.Any]) -> None:
        if value is not None:
            raise InvalidDirectiveError('The deprecated directive does not expect an expression')

        if self._is_deprecated:
            raise InvalidDirectiveError('Duplicated deprecated directive')

        if len(self._structs) > 1:
            raise InvalidDirectiveError('The deprecated directive cannot be placed in the response section')

        if not self._structs[-1].empty:
            raise InvalidDirectiveError('The deprecated directive must be placed before the first '
                                        'attribute definition')

        self._is_deprecated = True



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import typing
from . import serializable
from . import error


class BitLengthAnalysisError(error.InvalidDefinitionError):
    pass


class DataSchemaBuilder:
    def __init__(self) -> None:
        self._fields = []       # type: typing.List[serializable.Field]
        self._constants = []    # type: typing.List[serializable.Constant]
        self._is_union = False
        self._bit_length_computed_at_least_once = False

    def add_field(self, field: serializable.Field) -> None:
        if self.union and self._bit_length_computed_at_least_once:
            # Refer to the DSDL specification for the background information.
            raise BitLengthAnalysisError('Inter-field offset is not defined for unions; '
                                         'previously performed bit length analysis is invalid')
        assert isinstance(field, serializable.Field)
        self._fields.append(field)

    def add_constant(self, constant: serializable.Constant) -> None:
        assert isinstance(constant, serializable.Constant)
        self._constants.append(constant)

    @property
    def fields(self) -> typing.List[serializable.Field]:
        assert all(map(lambda x: isinstance(x, serializable.Field), self._fields))
        return self._fields

    @property
    def constants(self) -> typing.List[serializable.Constant]:
        assert all(map(lambda x: isinstance(x, serializable.Constant), self._constants))
        return self._constants

    @property
    def attributes(self) -> typing.List[serializable.Attribute]:  # noinspection PyTypeChecker
        out = []  # type: typing.List[serializable.Attribute]
        out += self.fields
        out += self.constants
        return out

    @property
    def empty(self) -> bool:
        return not self._fields and not self._constants

    @property
    def union(self) -> bool:
        return self._is_union

    def make_union(self) -> None:
        assert not self.union, 'This operation is not idempotent'
        self._is_union = True

    def compute_bit_length_values(self) -> typing.Set[int]:     # oh mypy, why are you so stupid
        # We set this flag in order to detect invalid reliance on the bit length estimates for unions:
        # we process definitions sequentially, statement-by-statement, so we can't know if there are going to be
        # extra fields added after the bit length values are computed. If we are building a regular structure,
        # this is fine, because in that case each computed value refers to the offset of the next field (if there
        # is one) or the total length of the structure (if we're past the last field). With unions, however,
        # there is no concept of inter-field offset because a union holds exactly one field at any moment;
        # only the total offset (i.e., total size) is defined.
        self._bit_length_computed_at_least_once = True

        field_type_gen = map(lambda f: f.data_type, self.fields)
        if self.union:
            out = serializable.compute_bit_length_values_for_tagged_union(field_type_gen)
        else:
            out = serializable.compute_bit_length_values_for_struct(field_type_gen)

        assert isinstance(out, set) and len(out) > 0
        return out



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import re
import enum
import string
import typing
import itertools
import fractions
from . import expression
from . import error
from . import port_id_ranges


BitLengthRange = typing.NamedTuple('BitLengthRange', [('min', int), ('max', int)])

ValueRange = typing.NamedTuple('ValueRange', [('min', fractions.Fraction), ('max', fractions.Fraction)])

Version = typing.NamedTuple('Version', [('major', int), ('minor', int)])


_VALID_FIRST_CHARACTERS_OF_NAME = string.ascii_letters + '_'
_VALID_CONTINUATION_CHARACTERS_OF_NAME = _VALID_FIRST_CHARACTERS_OF_NAME + string.digits

# Disallowed name patterns apply to any part of any name, e.g.,
# an attribute name, a namespace component, type name, etc.
# The pattern must produce an exact match to trigger a name error.
# All patterns are case-insensitive.
_DISALLOWED_NAME_PATTERNS = [
    r'truncated',
    r'saturated',
    r'true',
    r'false',
    r'bool',
    r'void\d*',
    r'u?int\d*',
    r'u?q\d+_\d+',
    r'float\d*',
    r'optional',
    r'aligned',
    r'const',
    r'struct',
    r'super',
    r'template',
    r'enum',
    r'self',
    r'and',
    r'or',
    r'not',
    r'auto',
    r'type',
    r'con',
    r'prn',
    r'aux',
    r'nul',
    r'com\d?',
    r'lpt\d?',
    r'_.*_',
]


class TypeParameterError(error.InvalidDefinitionError):
    pass


class InvalidBitLengthError(TypeParameterError):
    pass


class InvalidCastModeError(TypeParameterError):
    pass


class InvalidNumberOfElementsError(TypeParameterError):
    pass


class InvalidNameError(TypeParameterError):
    pass


class InvalidVersionError(TypeParameterError):
    pass


class InvalidConstantValueError(TypeParameterError):
    pass


class InvalidTypeError(TypeParameterError):
    pass


class AttributeNameCollision(TypeParameterError):
    pass


class InvalidFixedPortIDError(TypeParameterError):
    pass


class MalformedUnionError(TypeParameterError):
    pass


class DeprecatedDependencyError(TypeParameterError):
    pass


class SerializableType(expression.Any):
    """
    Invoking __str__() on a data type returns its uniform normalized definition, e.g.:
        - uavcan.node.Heartbeat.1.0[<=36]
        - truncated float16[<=36]
    """

    TYPE_NAME = 'metaserializable'

    @property
    def bit_length_range(self) -> BitLengthRange:
        # This default implementation uses BLV analysis, so it is very slow.
        # Derived classes can redefine it as appropriate.
        blv = self.compute_bit_length_values()
        return BitLengthRange(min(blv), max(blv))

    def compute_bit_length_values(self) -> typing.Set[int]:     # pragma: no cover
        """
        A set of all possible bit length values for the encoded representation of the data type.
        With complex data types, a full bit length set estimation may lead to a combinatorial explosion.
        This function must never return an empty set.
        The following invariants hold:
        >>> self.bit_length_range.min == min(self.compute_bit_length_values())
        >>> self.bit_length_range.max == max(self.compute_bit_length_values())
        """
        raise NotImplementedError

    def _attribute(self, name: expression.String) -> expression.Any:
        if name.native_value == '_bit_length_':
            try:
                return expression.Set(map(expression.Rational, self.compute_bit_length_values()))
            except TypeError:
                pass

        return super(SerializableType, self)._attribute(name)  # Hand over up the inheritance chain, important

    def __str__(self) -> str:   # pragma: no cover
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SerializableType):
            same_type = isinstance(other, type(self)) and isinstance(self, type(other))
            return same_type and str(self) == str(other)
        else:
            return NotImplemented


class PrimitiveType(SerializableType):
    MAX_BIT_LENGTH = 64

    class CastMode(enum.Enum):
        SATURATED = 0
        TRUNCATED = 1

    def __init__(self,
                 bit_length: int,
                 cast_mode: 'PrimitiveType.CastMode'):
        super(PrimitiveType, self).__init__()
        self._bit_length = int(bit_length)
        self._cast_mode = cast_mode

        if self._bit_length < 1:
            raise InvalidBitLengthError('Bit length must be positive')

        if self._bit_length > self.MAX_BIT_LENGTH:
            raise InvalidBitLengthError('Bit length cannot exceed %r' % self.MAX_BIT_LENGTH)

    @property
    def bit_length(self) -> int:
        """All primitives are of a fixed bit length, hence just one value is enough."""
        return self._bit_length

    def compute_bit_length_values(self) -> typing.Set[int]:
        return {self._bit_length}

    @property
    def cast_mode(self) -> 'PrimitiveType.CastMode':
        return self._cast_mode

    @property
    def _cast_mode_name(self) -> str:
        """For internal use only."""
        return {
            self.CastMode.SATURATED: 'saturated',
            self.CastMode.TRUNCATED: 'truncated',
        }[self.cast_mode]

    def __str__(self) -> str:   # pragma: no cover
        raise NotImplementedError

    def __repr__(self) -> str:
        return '%s(bit_length=%r, cast_mode=%r)' % (self.__class__.__name__, self.bit_length, self.cast_mode)


class BooleanType(PrimitiveType):
    def __init__(self, cast_mode: PrimitiveType.CastMode):
        super(BooleanType, self).__init__(bit_length=1, cast_mode=cast_mode)

        if cast_mode != PrimitiveType.CastMode.SATURATED:
            raise InvalidCastModeError('Invalid cast mode for boolean: %r' % cast_mode)

    def __str__(self) -> str:
        return self._cast_mode_name + ' bool'


class ArithmeticType(PrimitiveType):
    def __init__(self,
                 bit_length: int,
                 cast_mode: PrimitiveType.CastMode):
        super(ArithmeticType, self).__init__(bit_length, cast_mode)

    @property
    def inclusive_value_range(self) -> ValueRange:   # pragma: no cover
        raise NotImplementedError

    def __str__(self) -> str:   # pragma: no cover
        raise NotImplementedError


class IntegerType(ArithmeticType):
    def __init__(self,
                 bit_length: int,
                 cast_mode: PrimitiveType.CastMode):
        super(IntegerType, self).__init__(bit_length, cast_mode)

        if self._bit_length < 2:
            raise InvalidBitLengthError('Bit length of integer types cannot be less than 2')

    @property
    def inclusive_value_range(self) -> ValueRange:   # pragma: no cover
        raise NotImplementedError

    def __str__(self) -> str:   # pragma: no cover
        raise NotImplementedError


class SignedIntegerType(IntegerType):
    def __init__(self,
                 bit_length: int,
                 cast_mode: PrimitiveType.CastMode):
        super(SignedIntegerType, self).__init__(bit_length, cast_mode)

        if cast_mode != PrimitiveType.CastMode.SATURATED:
            raise InvalidCastModeError('Invalid cast mode for signed integer: %r' % cast_mode)

    @property
    def inclusive_value_range(self) -> ValueRange:
        uint_max_half = ((1 << self.bit_length) - 1) // 2
        return ValueRange(min=fractions.Fraction(-uint_max_half - 1),
                          max=fractions.Fraction(+uint_max_half))

    def __str__(self) -> str:
        return self._cast_mode_name + ' int' + str(self.bit_length)


class UnsignedIntegerType(IntegerType):
    def __init__(self,
                 bit_length: int,
                 cast_mode: PrimitiveType.CastMode):
        super(UnsignedIntegerType, self).__init__(bit_length, cast_mode)

    @property
    def inclusive_value_range(self) -> ValueRange:
        return ValueRange(min=fractions.Fraction(0),
                          max=fractions.Fraction((1 << self.bit_length) - 1))

    def __str__(self) -> str:
        return self._cast_mode_name + ' uint' + str(self.bit_length)


class FloatType(ArithmeticType):
    def __init__(self,
                 bit_length: int,
                 cast_mode: PrimitiveType.CastMode):
        super(FloatType, self).__init__(bit_length, cast_mode)

        try:
            self._magnitude = {
                16: fractions.Fraction(65504),
                32: fractions.Fraction('3.40282346638528859812e+38'),
                64: fractions.Fraction('1.79769313486231570815e+308'),
            }[self.bit_length]  # type: fractions.Fraction
        except KeyError:
            raise InvalidBitLengthError('Invalid bit length for float type: %d' % bit_length) from None

    @property
    def inclusive_value_range(self) -> ValueRange:
        return ValueRange(min=-self._magnitude,
                          max=+self._magnitude)

    def __str__(self) -> str:
        return self._cast_mode_name + ' float' + str(self.bit_length)


def _unittest_primitive() -> None:
    from pytest import raises

    assert str(BooleanType(PrimitiveType.CastMode.SATURATED)) == 'saturated bool'

    assert str(SignedIntegerType(15, PrimitiveType.CastMode.SATURATED)) == 'saturated int15'
    assert SignedIntegerType(64, PrimitiveType.CastMode.SATURATED).bit_length_range == (64, 64)
    assert SignedIntegerType(8, PrimitiveType.CastMode.SATURATED).inclusive_value_range == (-128, 127)

    assert str(UnsignedIntegerType(15, PrimitiveType.CastMode.TRUNCATED)) == 'truncated uint15'
    assert UnsignedIntegerType(53, PrimitiveType.CastMode.SATURATED).bit_length_range == (53, 53)
    assert UnsignedIntegerType(32, PrimitiveType.CastMode.SATURATED).inclusive_value_range == (0, 0xFFFFFFFF)

    assert str(FloatType(64, PrimitiveType.CastMode.SATURATED)) == 'saturated float64'
    assert FloatType(32, PrimitiveType.CastMode.SATURATED).bit_length_range == (32, 32)
    assert FloatType(16, PrimitiveType.CastMode.SATURATED).inclusive_value_range == (-65504, +65504)

    with raises(InvalidBitLengthError):
        FloatType(8, PrimitiveType.CastMode.TRUNCATED)

    with raises(InvalidBitLengthError):
        SignedIntegerType(1, PrimitiveType.CastMode.SATURATED)

    with raises(InvalidBitLengthError):
        SignedIntegerType(0, PrimitiveType.CastMode.SATURATED)

    with raises(InvalidBitLengthError):
        UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED)

    with raises(InvalidBitLengthError):
        UnsignedIntegerType(65, PrimitiveType.CastMode.TRUNCATED)

    assert repr(SignedIntegerType(24, PrimitiveType.CastMode.SATURATED)) == \
        'SignedIntegerType(bit_length=24, cast_mode=<CastMode.SATURATED: 0>)'

    a = UnsignedIntegerType(2, PrimitiveType.CastMode.TRUNCATED)
    b = BooleanType(PrimitiveType.CastMode.SATURATED)
    assert hash(a) != hash(b)
    assert hash(a) == hash(UnsignedIntegerType(2, PrimitiveType.CastMode.TRUNCATED))
    assert a == UnsignedIntegerType(2, PrimitiveType.CastMode.TRUNCATED)
    assert b != UnsignedIntegerType(2, PrimitiveType.CastMode.TRUNCATED)
    assert a != b
    assert b == BooleanType(PrimitiveType.CastMode.SATURATED)
    assert b != 123    # Not implemented


class VoidType(SerializableType):
    MAX_BIT_LENGTH = 64

    def __init__(self, bit_length: int):
        super(VoidType, self).__init__()
        self._bit_length = int(bit_length)

        if self._bit_length < 1:
            raise InvalidBitLengthError('Bit length must be positive')

        if self._bit_length > self.MAX_BIT_LENGTH:
            raise InvalidBitLengthError('Bit length cannot exceed %r' % self.MAX_BIT_LENGTH)

    @property
    def bit_length(self) -> int:
        return self._bit_length

    def compute_bit_length_values(self) -> typing.Set[int]:
        return {self._bit_length}

    def __str__(self) -> str:
        return 'void%d' % self.bit_length

    def __repr__(self) -> str:
        return 'VoidType(bit_length=%d)' % self.bit_length


def _unittest_void() -> None:
    from pytest import raises

    assert VoidType(1).bit_length_range == (1, 1)
    assert str(VoidType(13)) == 'void13'
    assert repr(VoidType(64)) == 'VoidType(bit_length=64)'
    assert VoidType(22).compute_bit_length_values() == {22}

    with raises(InvalidBitLengthError):
        VoidType(1)
        VoidType(0)

    with raises(InvalidBitLengthError):
        VoidType(64)
        VoidType(65)


class ArrayType(SerializableType):
    def __init__(self,
                 element_type: SerializableType,
                 capacity: int):
        super(ArrayType, self).__init__()
        self._element_type = element_type
        self._capacity = int(capacity)
        if self._capacity < 1:
            raise InvalidNumberOfElementsError('Array capacity cannot be less than 1')

    @property
    def element_type(self) -> SerializableType:
        return self._element_type

    @property
    def capacity(self) -> int:
        return self._capacity

    @property
    def string_like(self) -> bool:
        """
        Returns True if the array might contain a text string, in which case it is termed to be "string-like".
        A string-like array is a dynamic array of uint8.
        """
        return False

    def compute_bit_length_values(self) -> typing.Set[int]:     # pragma: no cover
        raise NotImplementedError

    def __str__(self) -> str:   # pragma: no cover
        raise NotImplementedError


class FixedLengthArrayType(ArrayType):
    def __init__(self,
                 element_type: SerializableType,
                 capacity: int):
        super(FixedLengthArrayType, self).__init__(element_type, capacity)

    @property
    def bit_length_range(self) -> BitLengthRange:
        return BitLengthRange(min=self.element_type.bit_length_range.min * self.capacity,
                              max=self.element_type.bit_length_range.max * self.capacity)

    def compute_bit_length_values(self) -> typing.Set[int]:
        # combinations_with_replacement() implements the standard n-combination function.
        combinations = itertools.combinations_with_replacement(self.element_type.compute_bit_length_values(),
                                                               self.capacity)
        # We do not care about permutations because the final bit length is invariant to the order of
        # serialized elements. Having found all combinations, we need to obtain a set of resulting length values.
        return set(map(sum, combinations))

    def __str__(self) -> str:
        return '%s[%d]' % (self.element_type, self.capacity)

    def __repr__(self) -> str:
        return 'FixedLengthArrayType(element_type=%r, capacity=%r)' % (self.element_type, self.capacity)


def _unittest_fixed_array() -> None:
    from pytest import raises

    su8 = UnsignedIntegerType(8, cast_mode=PrimitiveType.CastMode.TRUNCATED)
    ti64 = SignedIntegerType(64, cast_mode=PrimitiveType.CastMode.SATURATED)

    assert str(FixedLengthArrayType(su8, 4)) == 'truncated uint8[4]'
    assert str(FixedLengthArrayType(ti64, 1)) == 'saturated int64[1]'

    assert not FixedLengthArrayType(su8, 4).string_like
    assert not FixedLengthArrayType(ti64, 1).string_like

    assert FixedLengthArrayType(su8, 4).bit_length_range == (32, 32)
    assert FixedLengthArrayType(su8, 200).capacity == 200
    assert FixedLengthArrayType(ti64, 200).element_type is ti64

    with raises(InvalidNumberOfElementsError):
        FixedLengthArrayType(ti64, 0)

    assert repr(FixedLengthArrayType(ti64, 128)) == \
        'FixedLengthArrayType(element_type=SignedIntegerType(bit_length=64, cast_mode=<CastMode.SATURATED: 0>), ' \
        'capacity=128)'

    small = FixedLengthArrayType(su8, 2)
    assert small.compute_bit_length_values() == {16}


class VariableLengthArrayType(ArrayType):
    def __init__(self,
                 element_type: SerializableType,
                 capacity: int):
        super(VariableLengthArrayType, self).__init__(element_type, capacity)

    @property
    def string_like(self) -> bool:
        et = self.element_type      # Without this temporary MyPy yields a false positive type error
        return isinstance(et, UnsignedIntegerType) and (et.bit_length == 8)

    @property
    def length_prefix_bit_length(self) -> int:
        return self.capacity.bit_length()

    @property
    def bit_length_range(self) -> BitLengthRange:
        return BitLengthRange(
            min=self.length_prefix_bit_length,
            max=self.length_prefix_bit_length + self.element_type.bit_length_range.max * self.capacity
        )

    def compute_bit_length_values(self) -> typing.Set[int]:
        # Please refer to the corresponding implementation for the static array.
        # The idea here is that we treat the dynamic array as a combination of static arrays of different sizes,
        # from zero elements up to the maximum number of elements.
        output = set()           # type: typing.Set[int]
        for capacity in range(self.capacity + 1):
            combinations = itertools.combinations_with_replacement(self.element_type.compute_bit_length_values(),
                                                                   capacity)
            output |= set(map(lambda c: sum(c) + self.length_prefix_bit_length, combinations))

        return output

    def __str__(self) -> str:
        return '%s[<=%d]' % (self.element_type, self.capacity)

    def __repr__(self) -> str:
        return 'VariableLengthArrayType(element_type=%r, capacity=%r)' % (self.element_type, self.capacity)


def _unittest_variable_array() -> None:
    from pytest import raises

    tu8 = UnsignedIntegerType(8, cast_mode=PrimitiveType.CastMode.TRUNCATED)
    si64 = SignedIntegerType(64, cast_mode=PrimitiveType.CastMode.SATURATED)

    assert str(VariableLengthArrayType(tu8, 4)) == 'truncated uint8[<=4]'
    assert str(VariableLengthArrayType(si64, 255)) == 'saturated int64[<=255]'

    assert VariableLengthArrayType(tu8, 4).string_like
    assert not VariableLengthArrayType(si64, 1).string_like

    # Mind the length prefix!
    assert VariableLengthArrayType(tu8, 3).bit_length_range == (2, 26)
    assert VariableLengthArrayType(tu8, 1).bit_length_range == (1, 9)
    assert VariableLengthArrayType(tu8, 255).bit_length_range == (8, 2048)
    assert VariableLengthArrayType(tu8, 65535).bit_length_range == (16, 16 + 65535 * 8)

    assert VariableLengthArrayType(tu8, 200).capacity == 200
    assert VariableLengthArrayType(tu8, 200).element_type is tu8

    with raises(InvalidNumberOfElementsError):
        VariableLengthArrayType(si64, 0)

    assert repr(VariableLengthArrayType(si64, 128)) == \
        'VariableLengthArrayType(element_type=SignedIntegerType(bit_length=64, cast_mode=<CastMode.SATURATED: 0>), ' \
        'capacity=128)'

    # The following was computed manually; it is easy to validate:
    # we have zero, one, or two elements of 8 bits each; plus 2 bit wide tag; therefore:
    # {2 + 0, 2 + 8, 2 + 16}
    small = VariableLengthArrayType(tu8, 2)
    assert small.compute_bit_length_values() == {2, 10, 18}

    # This one gets a little tricky, so pull out a piece of paper an a pencil.
    # So the nested type, as defined above, has the following set: {2, 10, 18}.
    # We can have up to two elements of that type, so what we get can be expressed graphically as follows:
    #    A   B | +
    # ---------+------
    #    2   2 |  4
    #   10   2 | 12
    #   18   2 | 20
    #    2  10 | 12
    #   10  10 | 20
    #   18  10 | 28
    #    2  18 | 20
    #   10  18 | 28
    #   18  18 | 36
    #
    # If we were to remove duplicates, we end up with: {4, 12, 20, 28, 36}
    outer = FixedLengthArrayType(small, 2)
    assert outer.compute_bit_length_values() == {4, 12, 20, 28, 36}


class Attribute:    # TODO: should extend expression.Any to support advanced introspection/reflection.
    def __init__(self, data_type: SerializableType, name: str):
        self._data_type = data_type
        self._name = str(name)

        if isinstance(data_type, VoidType):
            if self._name:
                raise InvalidNameError('Void-typed fields can be used only for padding and cannot be named')
        else:
            _check_name(self._name)

    @property
    def data_type(self) -> SerializableType:
        return self._data_type

    @property
    def name(self) -> str:
        return self._name

    def __str__(self) -> str:
        return '%s %s' % (self.data_type, self.name)

    def __repr__(self) -> str:
        return '%s(data_type=%r, name=%r)' % (self.__class__.__name__, self.data_type, self.name)


class Field(Attribute):
    pass


class PaddingField(Field):
    def __init__(self, data_type: VoidType):
        if not isinstance(data_type, VoidType):
            raise TypeParameterError('Padding fields must be of the void type')

        super(PaddingField, self).__init__(data_type, '')


class Constant(Attribute):
    def __init__(self,
                 data_type: SerializableType,
                 name: str,
                 value: expression.Any):
        super(Constant, self).__init__(data_type, name)

        if not isinstance(value, expression.Primitive):
            raise InvalidConstantValueError('The constant value must be a primitive expression value')

        self._value = value
        del value

        # Interestingly, both the type of the constant and its value are instances of the same meta-type: expression.
        # BooleanType inherits from expression.Any, same as expression.Boolean.
        if isinstance(data_type, BooleanType):      # Boolean constant
            if not isinstance(self._value, expression.Boolean):
                raise InvalidConstantValueError('Invalid value for boolean constant: %r' % self._value)

        elif isinstance(data_type, IntegerType):    # Integer constant
            if isinstance(self._value, expression.Rational):
                if not self._value.is_integer():
                    raise InvalidConstantValueError('The value of an integer constant must be an integer; got %s' %
                                                    self._value)
            elif isinstance(self._value, expression.String):
                as_bytes = self._value.native_value.encode('utf8')
                if len(as_bytes) != 1:
                    raise InvalidConstantValueError('A constant string must be exactly one ASCII character long')

                if not isinstance(data_type, UnsignedIntegerType) or data_type.bit_length != 8:
                    raise InvalidConstantValueError('Constant strings can be used only with uint8')

                self._value = expression.Rational(ord(as_bytes))    # Replace string with integer
            else:
                raise InvalidConstantValueError('Invalid value type for integer constant: %r' % self._value)

        elif isinstance(data_type, FloatType):      # Floating point constant
            if not isinstance(self._value, expression.Rational):
                raise InvalidConstantValueError('Invalid value type for float constant: %r' % self._value)

        else:
            raise InvalidTypeError('Invalid constant type: %r' % data_type)

        assert isinstance(self._value, expression.Any)
        assert isinstance(self._value, expression.Rational) == isinstance(self.data_type, (FloatType, IntegerType))
        assert isinstance(self._value, expression.Boolean)  == isinstance(self.data_type, BooleanType)

        # Range check
        if isinstance(self._value, expression.Rational):
            assert isinstance(data_type, ArithmeticType)
            rng = data_type.inclusive_value_range
            if not (rng.min <= self._value.native_value <= rng.max):
                raise InvalidConstantValueError('Constant value %s exceeds the range of its data type %s' %
                                                (self._value, data_type))

    @property
    def value(self) -> expression.Any:
        return self._value

    def __str__(self) -> str:
        return '%s %s = %s' % (self.data_type, self.name, self.value)

    def __repr__(self) -> str:
        return 'Constant(data_type=%r, name=%r, value=%r)' % (self.data_type, self.name, self._value)


def _unittest_attribute() -> None:
    from pytest import raises

    assert str(Field(BooleanType(PrimitiveType.CastMode.SATURATED), 'flag')) == 'saturated bool flag'
    assert repr(Field(BooleanType(PrimitiveType.CastMode.SATURATED), 'flag')) == \
        'Field(data_type=BooleanType(bit_length=1, cast_mode=<CastMode.SATURATED: 0>), name=\'flag\')'

    assert str(PaddingField(VoidType(32))) == 'void32 '     # Mind the space!
    assert repr(PaddingField(VoidType(1))) == 'PaddingField(data_type=VoidType(bit_length=1), name=\'\')'

    with raises(TypeParameterError, match='.*void.*'):
        # noinspection PyTypeChecker
        repr(PaddingField(SignedIntegerType(8, PrimitiveType.CastMode.SATURATED)))   # type: ignore

    data_type = SignedIntegerType(32, PrimitiveType.CastMode.SATURATED)
    const = Constant(data_type, 'FOO_CONST', expression.Rational(-123))
    assert str(const) == 'saturated int32 FOO_CONST = -123'
    assert const.data_type is data_type
    assert const.name == 'FOO_CONST'
    assert const.value == expression.Rational(-123)

    assert repr(const) == 'Constant(data_type=%r, name=\'FOO_CONST\', value=rational(-123))' % data_type


class CompositeType(SerializableType):
    MAX_NAME_LENGTH = 63
    MAX_VERSION_NUMBER = 255
    NAME_COMPONENT_SEPARATOR = '.'

    def __init__(self,
                 name:          str,
                 version:       Version,
                 attributes:    typing.Iterable[Attribute],
                 deprecated:    bool,
                 fixed_port_id: typing.Optional[int],
                 source_file_path:  str):
        self._name = str(name).strip()
        self._version = version
        self._attributes = list(attributes)
        self._deprecated = bool(deprecated)
        self._fixed_port_id = None if fixed_port_id is None else int(fixed_port_id)
        self._source_file_path = str(source_file_path)

        # Name check
        if not self._name:
            raise InvalidNameError('Composite type name cannot be empty')

        if self.NAME_COMPONENT_SEPARATOR not in self._name:
            raise InvalidNameError('Root namespace is not specified')

        if len(self._name) > self.MAX_NAME_LENGTH:
            raise InvalidNameError('Name is too long: %r is longer than %d characters' %
                                   (self._name, self.MAX_NAME_LENGTH))

        for component in self._name.split(self.NAME_COMPONENT_SEPARATOR):
            _check_name(component)

        # Version check
        version_valid = (0 <= self._version.major <= self.MAX_VERSION_NUMBER) and\
                        (0 <= self._version.minor <= self.MAX_VERSION_NUMBER) and\
                        ((self._version.major + self._version.minor) > 0)

        if not version_valid:
            raise InvalidVersionError('Invalid version numbers: %s.%s' % (self._version.major, self._version.minor))

        # Attribute check
        used_names = set()      # type: typing.Set[str]
        for a in self._attributes:
            if a.name and a.name in used_names:
                raise AttributeNameCollision('Multiple attributes under the same name: %r' % a.name)
            else:
                used_names.add(a.name)

        # Port ID check
        port_id = self._fixed_port_id
        if port_id is not None:
            assert port_id is not None
            if isinstance(self, ServiceType):
                if not (0 <= port_id <= port_id_ranges.MAX_SERVICE_ID):
                    raise InvalidFixedPortIDError('Fixed service ID %r is not valid' % port_id)
            else:
                if not (0 <= port_id <= port_id_ranges.MAX_SUBJECT_ID):
                    raise InvalidFixedPortIDError('Fixed subject ID %r is not valid' % port_id)

        # Consistent deprecation check.
        # A non-deprecated type cannot be dependent on deprecated types.
        # A deprecated type can be dependent on anything.
        if not self.deprecated:
            for a in self._attributes:
                t = a.data_type
                if isinstance(t, CompositeType):
                    if t.deprecated:
                        raise DeprecatedDependencyError('A type cannot depend on deprecated types '
                                                        'unless it is also deprecated.')

    def is_mutually_bit_compatible_with(self, other: 'CompositeType') -> bool:
        """
        Checks for bit compatibility between two data types.
        The current implementation uses a relaxed simplified check that may yield a false-negative,
        but never a false-positive; i.e., it may fail to detect an incompatibility, but it is guaranteed
        to never report two data types as incompatible if they are compatible.
        The implementation may be updated in the future to use a strict check as defined in the specification
        while keeping the same API, so beware.
        """
        return self.compute_bit_length_values() == other.compute_bit_length_values()

    @property
    def full_name(self) -> str:
        """The full name, e.g., uavcan.node.Heartbeat"""
        return self._name

    @property
    def name_components(self) -> typing.List[str]:
        """Components of the full name as a list, e.g., ['uavcan', 'node', 'Heartbeat']"""
        return self._name.split(CompositeType.NAME_COMPONENT_SEPARATOR)

    @property
    def short_name(self) -> str:
        """The last component of the full name, e.g., Heartbeat of uavcan.node.Heartbeat"""
        return self.name_components[-1]

    @property
    def full_namespace(self) -> str:
        """The full name without the short name, e.g., uavcan.node for uavcan.node.Heartbeat"""
        return str(CompositeType.NAME_COMPONENT_SEPARATOR.join(self.name_components[:-1]))

    @property
    def root_namespace(self) -> str:
        """The first component of the full name, e.g., uavcan of uavcan.node.Heartbeat"""
        return self.name_components[0]

    @property
    def version(self) -> Version:
        return self._version

    @property
    def deprecated(self) -> bool:
        return self._deprecated

    @property
    def attributes(self) -> typing.List[Attribute]:
        return self._attributes[:]  # Return copy to prevent mutation

    @property
    def fields(self) -> typing.List[Field]:
        return [a for a in self.attributes if isinstance(a, Field)]

    @property
    def constants(self) -> typing.List[Constant]:
        return [a for a in self.attributes if isinstance(a, Constant)]

    @property
    def fixed_port_id(self) -> typing.Optional[int]:
        return self._fixed_port_id

    @property
    def has_fixed_port_id(self) -> bool:
        return self.fixed_port_id is not None

    @property
    def source_file_path(self) -> str:
        """Empty if this is a synthesized type, e.g. a service request or response section."""
        return self._source_file_path

    def compute_bit_length_values(self) -> typing.Set[int]:     # pragma: no cover
        raise NotImplementedError

    def _attribute(self, name: expression.String) -> expression.Any:
        """
        This is the handler for DSDL expressions like uavcan.node.Heartbeat.1.0.MODE_OPERATIONAL.
        """
        for c in self.constants:
            if c.name == name.native_value:
                assert isinstance(c.value, expression.Any)
                return c.value

        return super(CompositeType, self)._attribute(name)  # Hand over up the inheritance chain, this is important

    def __str__(self) -> str:
        return '%s.%d.%d' % (self.full_name, self.version.major, self.version.minor)

    def __repr__(self) -> str:
        return '%s(name=%r, version=%r, fields=%r, constants=%r, deprecated=%r, fixed_port_id=%r)' % \
            (self.__class__.__name__,
             self.full_name,
             self.version,
             self.fields,
             self.constants,
             self.deprecated,
             self.fixed_port_id)


class UnionType(CompositeType):
    MIN_NUMBER_OF_VARIANTS = 2

    def __init__(self,
                 name:             str,
                 version:          Version,
                 attributes:       typing.Iterable[Attribute],
                 deprecated:       bool,
                 fixed_port_id:    typing.Optional[int],
                 source_file_path: str):
        # Proxy all parameters directly to the base type - I wish we could do that
        # with kwargs while preserving the type information
        super(UnionType, self).__init__(name=name,
                                        version=version,
                                        attributes=attributes,
                                        deprecated=deprecated,
                                        fixed_port_id=fixed_port_id,
                                        source_file_path=source_file_path)

        if self.number_of_variants < self.MIN_NUMBER_OF_VARIANTS:
            raise MalformedUnionError('A tagged union cannot contain less than %d variants' %
                                      self.MIN_NUMBER_OF_VARIANTS)

        for a in attributes:
            if isinstance(a, PaddingField) or not a.name or isinstance(a.data_type, VoidType):
                raise MalformedUnionError('Padding fields not allowed in unions')

    @property
    def number_of_variants(self) -> int:
        return len(self.fields)

    @property
    def tag_bit_length(self) -> int:
        return (self.number_of_variants - 1).bit_length()

    @property
    def bit_length_range(self) -> BitLengthRange:
        blr = [f.data_type.bit_length_range for f in self.fields]
        return BitLengthRange(min=self.tag_bit_length + min([b.min for b in blr]),
                              max=self.tag_bit_length + max([b.max for b in blr]))

    def compute_bit_length_values(self) -> typing.Set[int]:
        return compute_bit_length_values_for_tagged_union(map(lambda f: f.data_type, self.fields))


class StructureType(CompositeType):
    @property
    def bit_length_range(self) -> BitLengthRange:
        blr = [f.data_type.bit_length_range for f in self.fields]
        return BitLengthRange(min=sum([b.min for b in blr]),
                              max=sum([b.max for b in blr]))

    def compute_bit_length_values(self) -> typing.Set[int]:
        return compute_bit_length_values_for_struct(map(lambda f: f.data_type, self.fields))


class ServiceType(CompositeType):
    def __init__(self,
                 name:                str,
                 version:             Version,
                 request_attributes:  typing.Iterable[Attribute],
                 response_attributes: typing.Iterable[Attribute],
                 request_is_union:    bool,
                 response_is_union:   bool,
                 deprecated:          bool,
                 fixed_port_id:       typing.Optional[int],
                 source_file_path:    str):
        request_meta_type = UnionType if request_is_union else StructureType  # type: type
        self._request_type = request_meta_type(name=name + '.Request',
                                               version=version,
                                               attributes=request_attributes,
                                               deprecated=deprecated,
                                               fixed_port_id=None,
                                               source_file_path='')  # type: CompositeType

        response_meta_type = UnionType if response_is_union else StructureType  # type: type
        self._response_type = response_meta_type(name=name + '.Response',
                                                 version=version,
                                                 attributes=response_attributes,
                                                 deprecated=deprecated,
                                                 fixed_port_id=None,
                                                 source_file_path='')  # type: CompositeType

        container_attributes = [
            Field(data_type=self._request_type,  name='request'),
            Field(data_type=self._response_type, name='response'),
        ]

        super(ServiceType, self).__init__(name=name,
                                          version=version,
                                          attributes=container_attributes,
                                          deprecated=deprecated,
                                          fixed_port_id=fixed_port_id,
                                          source_file_path=source_file_path)

    @property
    def request_type(self) -> CompositeType:
        return self._request_type

    @property
    def response_type(self) -> CompositeType:
        return self._response_type

    def compute_bit_length_values(self) -> typing.Set[int]:     # pragma: no cover
        raise TypeError('Service types are not directly serializable. Use either request or response.')


def _check_name(name: str) -> None:
    if not name:
        raise InvalidNameError('Name or namespace component cannot be empty')

    if name[0] not in _VALID_FIRST_CHARACTERS_OF_NAME:
        raise InvalidNameError('Name or namespace component cannot start with %r' % name[0])

    for char in name:
        if char not in _VALID_CONTINUATION_CHARACTERS_OF_NAME:
            raise InvalidNameError('Name or namespace component cannot contain %r' % char)

    for pat in _DISALLOWED_NAME_PATTERNS:
        if re.match(pat + '$', name, flags=re.IGNORECASE):
            raise InvalidNameError('Disallowed name: %r matches the following pattern: %s' % (name, pat))


def _unittest_composite_types() -> None:
    from pytest import raises

    def try_name(name: str) -> CompositeType:
        return CompositeType(name=name,
                             version=Version(0, 1),
                             attributes=[],
                             deprecated=False,
                             fixed_port_id=None,
                             source_file_path='')

    with raises(InvalidNameError, match='(?i).*empty.*'):
        try_name('')

    with raises(InvalidNameError, match='(?i).*root namespace.*'):
        try_name('T')

    with raises(InvalidNameError, match='(?i).*long.*'):
        try_name('namespace.another.deeper.' * 10 + 'LongTypeName')

    with raises(InvalidNameError, match='(?i).*component.*empty.*'):
        try_name('namespace.ns..T')

    with raises(InvalidNameError, match='(?i).*component.*empty.*'):
        try_name('.namespace.ns.T')

    with raises(InvalidNameError, match='(?i).*cannot start with.*'):
        try_name('namespace.0ns.T')

    with raises(InvalidNameError, match='(?i).*cannot start with.*'):
        try_name('namespace.ns.0T')

    with raises(InvalidNameError, match='(?i).*cannot contain.*'):
        try_name('namespace.n-s.T')

    assert try_name('root.nested.T').full_name == 'root.nested.T'
    assert try_name('root.nested.T').full_namespace == 'root.nested'
    assert try_name('root.nested.T').root_namespace == 'root'
    assert try_name('root.nested.T').short_name == 'T'

    with raises(MalformedUnionError, match='.*variants.*'):
        UnionType(name='a.A',
                  version=Version(0, 1),
                  attributes=[],
                  deprecated=False,
                  fixed_port_id=None,
                  source_file_path='')

    with raises(MalformedUnionError, match='(?i).*padding.*'):
        UnionType(name='a.A',
                  version=Version(0, 1),
                  attributes=[
                      Field(UnsignedIntegerType(16, PrimitiveType.CastMode.TRUNCATED), 'a'),
                      Field(SignedIntegerType(16, PrimitiveType.CastMode.SATURATED), 'b'),
                      PaddingField(VoidType(16)),
                  ],
                  deprecated=False,
                  fixed_port_id=None,
                  source_file_path='')

    _check_name('abc')
    _check_name('_abc')
    _check_name('abc_')
    _check_name('abc0')

    with raises(InvalidNameError):
        _check_name('0abc')

    with raises(InvalidNameError):
        _check_name('_abc_')

    with raises(InvalidNameError):
        _check_name('a-bc')

    with raises(InvalidNameError):
        _check_name('')

    with raises(InvalidNameError):
        _check_name('truncated')

    with raises(InvalidNameError):
        _check_name('COM1')

    with raises(InvalidNameError):
        _check_name('Aux')

    with raises(InvalidNameError):
        _check_name('float128')

    with raises(InvalidNameError):
        _check_name('q16_8')

    with raises(InvalidNameError):
        _check_name('uq1_32')

    def try_union_fields(field_types: typing.List[SerializableType]) -> UnionType:
        atr = []
        for i, t in enumerate(field_types):
            atr.append(Field(t, '_%d' % i))

        return UnionType(name='a.A',
                         version=Version(0, 1),
                         attributes=atr,
                         deprecated=False,
                         fixed_port_id=None,
                         source_file_path='')

    assert try_union_fields([
        UnsignedIntegerType(16, PrimitiveType.CastMode.TRUNCATED),
        SignedIntegerType(16, PrimitiveType.CastMode.SATURATED),
    ]).compute_bit_length_values() == {17}

    # The reference values for the following test are explained in the array tests above
    tu8 = UnsignedIntegerType(8, cast_mode=PrimitiveType.CastMode.TRUNCATED)
    small = VariableLengthArrayType(tu8, 2)
    outer = FixedLengthArrayType(small, 2)   # bit length values: {4, 12, 20, 28, 36}

    # Above plus one bit to each, plus 16-bit for the unsigned integer field
    assert try_union_fields([
        outer,
        SignedIntegerType(16, PrimitiveType.CastMode.SATURATED),
    ]).compute_bit_length_values() == {5, 13, 17, 21, 29, 37}

    def try_struct_fields(field_types: typing.List[SerializableType]) -> StructureType:
        atr = []
        for i, t in enumerate(field_types):
            atr.append(Field(t, '_%d' % i))

        return StructureType(name='a.A',
                             version=Version(0, 1),
                             attributes=atr,
                             deprecated=False,
                             fixed_port_id=None,
                             source_file_path='')

    assert try_struct_fields([
        UnsignedIntegerType(16, PrimitiveType.CastMode.TRUNCATED),
        SignedIntegerType(16, PrimitiveType.CastMode.SATURATED),
    ]).compute_bit_length_values() == {32}

    assert try_struct_fields([]).compute_bit_length_values() == {0}   # Empty sets forbidden

    assert try_struct_fields([
        outer,
        SignedIntegerType(16, PrimitiveType.CastMode.SATURATED),
    ]).compute_bit_length_values() == {4 + 16, 12 + 16, 20 + 16, 28 + 16, 36 + 16}

    assert try_struct_fields([outer]).compute_bit_length_values() == {4, 12, 20, 28, 36}


def compute_bit_length_values_for_struct(member_data_types: typing.Iterable['SerializableType']) -> typing.Set[int]:
    # As far as bit length combinations are concerned, structures are similar to static arrays.
    # Please refer to the bit length computation method for static arrays for reference.
    # The difference here is that the length value sets are not homogeneous across fields, as they
    # can be of different types, which sets structures apart from arrays. So instead of looking for
    # k-combinations, we need to find a Cartesian product of bit length value sets of each field.
    # For large structures with dynamic arrays this can be very computationally expensive.
    blv_sets = [x.compute_bit_length_values() for x in member_data_types]
    combinations = itertools.product(*blv_sets)

    # The interface prohibits empty sets at the output
    out = set(map(sum, combinations)) or {0}

    assert out and all(map(lambda x: isinstance(x, int) and x >= 0, out))
    return out      # type: ignore


def compute_bit_length_values_for_tagged_union(member_data_types: typing.Iterable[SerializableType]) -> typing.Set[int]:
    ts = list(member_data_types)
    del member_data_types

    if len(ts) < UnionType.MIN_NUMBER_OF_VARIANTS:
        raise MalformedUnionError('Cannot perform bit length analysis on less than {0} members because '
                                  'tagged unions are not defined for less than {0} variants'
                                  .format(UnionType.MIN_NUMBER_OF_VARIANTS))

    tag_bit_length = (len(ts) - 1).bit_length()
    assert tag_bit_length > 0

    # Unions are easy to handle because when serialized, a union is essentially just a single field,
    # prefixed with a fixed-length integer tag. So we just build a full set of combinations and then
    # add the tag length to each element. Easy.
    combinations = set()                                        # type: typing.Set[int]
    for t in ts:
        combinations |= t.compute_bit_length_values()

    out = set(map(lambda c: tag_bit_length + c, combinations))

    assert out and all(map(lambda x: isinstance(x, int) and x > 0, out))
    return out



from six import text_type, python_2_unicode_compatible

from parsimonious.utils import StrAndRepr


@python_2_unicode_compatible
class ParseError(StrAndRepr, Exception):
    """A call to ``Expression.parse()`` or ``match()`` didn't match."""

    def __init__(self, text, pos=-1, expr=None):
        # It would be nice to use self.args, but I don't want to pay a penalty
        # to call descriptors or have the confusion of numerical indices in
        # Expression.match_core().
        self.text = text
        self.pos = pos
        self.expr = expr

    def __str__(self):
        rule_name = ((u"'%s'" % self.expr.name) if self.expr.name else
                     text_type(self.expr))
        return u"Rule %s didn't match at '%s' (line %s, column %s)." % (
                rule_name,
                self.text[self.pos:self.pos + 20],
                self.line(),
                self.column())

    # TODO: Add line, col, and separated-out error message so callers can build
    # their own presentation.

    def line(self):
        """Return the 1-based line number where the expression ceased to
        match."""
        # This is a method rather than a property in case we ever wanted to
        # pass in which line endings we want to use.
        return self.text.count('\n', 0, self.pos) + 1

    def column(self):
        """Return the 1-based column where the expression ceased to match."""
        # We choose 1-based because that's what Python does with SyntaxErrors.
        try:
            return self.pos - self.text.rindex('\n', 0, self.pos)
        except ValueError:
            return self.pos + 1


@python_2_unicode_compatible
class IncompleteParseError(ParseError):
    """A call to ``parse()`` matched a whole Expression but did not consume the
    entire text."""

    def __str__(self):
        return u"Rule '%s' matched in its entirety, but it didn't consume all the text. The non-matching portion of the text begins with '%s' (line %s, column %s)." % (
                self.expr.name,
                self.text[self.pos:self.pos + 20],
                self.line(),
                self.column())


class VisitationError(Exception):
    """Something went wrong while traversing a parse tree.

    This exception exists to augment an underlying exception with information
    about where in the parse tree the error occurred. Otherwise, it could be
    tiresome to figure out what went wrong; you'd have to play back the whole
    tree traversal in your head.

    """
    # TODO: Make sure this is pickleable. Probably use @property pattern. Make
    # the original exc and node available on it if they don't cause a whole
    # raft of stack frames to be retained.
    def __init__(self, exc, exc_class, node):
        """Construct.

        :arg exc: What went wrong. We wrap this and add more info.
        :arg node: The node at which the error occurred

        """
        self.original_class = exc_class
        super(VisitationError, self).__init__(
            '%s: %s\n\n'
            'Parse tree:\n'
            '%s' %
            (exc_class.__name__,
             exc,
             node.prettily(error=node)))


class BadGrammar(StrAndRepr, Exception):
    """Something was wrong with the definition of a grammar.

    Note that a ParseError might be raised instead if the error is in the
    grammar definition syntax.

    """


@python_2_unicode_compatible
class UndefinedLabel(BadGrammar):
    """A rule referenced in a grammar was never defined.

    Circular references and forward references are okay, but you have to define
    stuff at some point.

    """
    def __init__(self, label):
        self.label = label

    def __str__(self):
        return u'The label "%s" was never defined.' % self.label



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import typing
import operator
import functools
import fractions
import unicodedata
from . import error


OperatorOutput    = typing.TypeVar('OperatorOutput')
UnaryOperator     = typing.Callable[['Any'], OperatorOutput]
BinaryOperator    = typing.Callable[['Any', 'Any'], OperatorOutput]
AttributeOperator = typing.Callable[['Any', typing.Union['String', str]], OperatorOutput]


class InvalidOperandError(error.InvalidDefinitionError):
    pass


class UndefinedOperatorError(InvalidOperandError):
    """Thrown when there is no matching operator for the supplied arguments."""
    def __init__(self) -> None:
        super(UndefinedOperatorError, self).__init__('The requested operator is not defined for the provided arguments')


class UndefinedAttributeError(InvalidOperandError):
    """Thrown when the requested attribute does not exist."""
    def __init__(self) -> None:
        super(UndefinedAttributeError, self).__init__('Invalid attribute name')


class Any:
    """
    This abstract class represents an arbitrary intrinsic DSDL expression value.
    """
    # This attribute must be specified in the derived classes.
    # It contains the name of the data type implemented by the class.
    TYPE_NAME = None    # type: str

    def __hash__(self) -> int:
        raise NotImplementedError  # pragma: no cover

    def __eq__(self, other: object) -> bool:
        raise NotImplementedError  # pragma: no cover

    def __str__(self) -> str:
        """Must return a DSDL spec-compatible textual representation of the contained value suitable for printing."""
        raise NotImplementedError  # pragma: no cover

    def __repr__(self) -> str:
        return self.TYPE_NAME + '(' + str(self) + ')'

    #
    # Unary operators.
    #
    def _logical_not(self) -> 'Boolean': raise UndefinedOperatorError

    def _positive(self) -> 'Any': raise UndefinedOperatorError

    def _negative(self) -> 'Any': raise UndefinedOperatorError

    #
    # Binary operators.
    # The types of the operators defined here must match the specification.
    # Make sure to use least generic types in the derived classes - Python allows covariant return types.
    #
    def _logical_or(self, right: 'Any')  -> 'Boolean': raise UndefinedOperatorError
    def _logical_and(self, right: 'Any') -> 'Boolean': raise UndefinedOperatorError

    def _equal(self, right: 'Any')            -> 'Boolean': raise UndefinedOperatorError  # pragma: no branch
    def _less_or_equal(self, right: 'Any')    -> 'Boolean': raise UndefinedOperatorError
    def _greater_or_equal(self, right: 'Any') -> 'Boolean': raise UndefinedOperatorError
    def _less(self, right: 'Any')             -> 'Boolean': raise UndefinedOperatorError
    def _greater(self, right: 'Any')          -> 'Boolean': raise UndefinedOperatorError

    def _bitwise_or(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _bitwise_or_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _bitwise_xor(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _bitwise_xor_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _bitwise_and(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _bitwise_and_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _add(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _add_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _subtract(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _subtract_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _multiply(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _multiply_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _divide(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _divide_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _modulo(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _modulo_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _power(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _power_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    #
    # Attribute access operator. It is a binary operator as well, but its semantics is slightly different.
    # Implementations must invoke super()._attribute() when they encounter an unknown attribute, to allow
    # the parent classes to handle the requested attribute as a fallback option.
    #
    def _attribute(self, name: 'String') -> 'Any': raise UndefinedAttributeError


# noinspection PyAbstractClass
class Primitive(Any):
    @property
    def native_value(self) -> typing.Any:
        raise NotImplementedError  # pragma: no cover


class Boolean(Primitive):
    TYPE_NAME = 'bool'

    def __init__(self, value: bool = False):
        if not isinstance(value, bool):
            raise ValueError('Cannot construct a Boolean instance from ' + type(value).__name__)

        self._value = value  # type: bool

    @property
    def native_value(self) -> bool:
        return self._value

    def __hash__(self) -> int:
        return int(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Boolean):
            return self._value == other._value
        else:  # pragma: no cover
            return NotImplemented

    def __str__(self) -> str:
        return 'true' if self._value else 'false'

    def __bool__(self) -> bool:     # For use in expressions without accessing "native_value"
        return self._value

    def _logical_not(self) -> 'Boolean':
        return Boolean(not self._value)

    def _logical_and(self, right: 'Any') -> 'Boolean':
        if isinstance(right, Boolean):
            return Boolean(self._value and right._value)
        else:
            raise UndefinedOperatorError

    def _logical_or(self, right: 'Any') -> 'Boolean':
        if isinstance(right, Boolean):
            return Boolean(self._value or right._value)
        else:
            raise UndefinedOperatorError

    def _equal(self, right: 'Any') -> 'Boolean':
        if isinstance(right, Boolean):
            return Boolean(self._value == right._value)
        else:
            raise UndefinedOperatorError


class Rational(Primitive):
    TYPE_NAME = 'rational'

    def __init__(self, value: typing.Union[int, fractions.Fraction]):
        # We must support float as well, because some operators on Fraction sometimes yield float, e.g. power.
        if not isinstance(value, (int, float, fractions.Fraction)):
            raise ValueError('Cannot construct a Rational instance from ' + type(value).__name__)
        self._value = fractions.Fraction(value)  # type: fractions.Fraction

    @property
    def native_value(self) -> fractions.Fraction:
        return self._value

    def as_native_integer(self) -> int:
        """
        Returns the inferior as a native integer,
        unless it cannot be represented as such without the loss of precision; i.e., if denominator != 1.
        """
        if self.is_integer():
            return self._value.numerator
        else:
            raise InvalidOperandError('Rational %s is not an integer' % self._value)

    def is_integer(self) -> bool:
        return self._value.denominator == 1

    def __hash__(self) -> int:
        return hash(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Rational):
            return self._value == other._value
        else:  # pragma: no cover
            return NotImplemented

    def __str__(self) -> str:
        return str(self._value)

    #
    # Unary operators.
    #
    def _positive(self) -> 'Rational':
        return Rational(+self._value)

    def _negative(self) -> 'Rational':
        return Rational(-self._value)

    #
    # Binary comparison operators.
    #
    def _generic_compare(self, right: 'Any', impl: typing.Callable[[typing.Any, typing.Any], bool]) -> Boolean:
        if isinstance(right, Rational):
            return Boolean(impl(self._value, right._value))
        else:
            raise UndefinedOperatorError

    def _equal(self, right: 'Any') -> 'Boolean':
        return self._generic_compare(right, operator.eq)

    def _less_or_equal(self, right: 'Any') -> 'Boolean':
        return self._generic_compare(right, operator.le)

    def _greater_or_equal(self, right: 'Any') -> 'Boolean':
        return self._generic_compare(right, operator.ge)

    def _less(self, right: 'Any') -> 'Boolean':
        return self._generic_compare(right, operator.lt)

    def _greater(self, right: 'Any') -> 'Boolean':
        return self._generic_compare(right, operator.gt)

    #
    # Binary bitwise operators.
    #
    def _generic_bitwise(self, right: 'Any', impl: typing.Callable[[typing.Any, typing.Any], typing.Any]) -> 'Rational':
        if isinstance(right, Rational):
            return Rational(impl(self.as_native_integer(), right.as_native_integer()))    # Throws if not an integer.
        else:
            raise UndefinedOperatorError

    def _bitwise_or(self, right: 'Any') -> 'Rational':
        return self._generic_bitwise(right, operator.or_)

    def _bitwise_xor(self, right: 'Any') -> 'Rational':
        return self._generic_bitwise(right, operator.xor)

    def _bitwise_and(self, right: 'Any') -> 'Rational':
        return self._generic_bitwise(right, operator.and_)

    #
    # Binary arithmetic operators.
    #
    def _generic_arithmetic(self,
                            right: 'Any',
                            impl: typing.Callable[[typing.Any, typing.Any], typing.Any]) -> 'Rational':
        if isinstance(right, Rational):
            try:
                result = impl(self._value, right._value)
            except ZeroDivisionError:
                raise InvalidOperandError('Cannot divide %s by zero' % self._value)
            else:
                return Rational(result)
        else:
            raise UndefinedOperatorError

    def _add(self, right: 'Any') -> 'Rational':
        return self._generic_arithmetic(right, operator.add)

    def _subtract(self, right: 'Any') -> 'Rational':
        return self._generic_arithmetic(right, operator.sub)

    def _multiply(self, right: 'Any') -> 'Rational':
        return self._generic_arithmetic(right, operator.mul)

    def _divide(self, right: 'Any') -> 'Rational':
        return self._generic_arithmetic(right, operator.truediv)

    def _modulo(self, right: 'Any') -> 'Rational':
        return self._generic_arithmetic(right, operator.mod)

    def _power(self, right: 'Any') -> 'Rational':
        return self._generic_arithmetic(right, operator.pow)


class String(Primitive):
    TYPE_NAME = 'string'

    def __init__(self, value: str):
        if not isinstance(value, str):
            raise ValueError('Cannot construct a String instance from ' + type(value).__name__)
        self._value = value  # type: str

    @property
    def native_value(self) -> str:
        return self._value

    def __hash__(self) -> int:
        return hash(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, String):
            return self._value == other._value
        else:  # pragma: no cover
            return NotImplemented

    def __str__(self) -> str:
        return repr(self._value)

    def _add(self, right: 'Any') -> 'String':
        if isinstance(right, String):
            return String(self._value + right._value)
        else:
            raise UndefinedOperatorError

    def _equal(self, right: 'Any') -> Boolean:
        if isinstance(right, String):
            def normalized(s: str) -> str:
                return unicodedata.normalize('NFC', s)

            return Boolean(normalized(self._value) == normalized(right._value))
        else:
            raise UndefinedOperatorError


# noinspection PyAbstractClass
class Container(Any):
    @property
    def element_type(self) -> typing.Type[Any]:
        raise NotImplementedError  # pragma: no cover

    def __iter__(self) -> typing.Iterator[typing.Any]:
        raise NotImplementedError  # pragma: no cover


class Set(Container):
    TYPE_NAME = 'set'

    # noinspection PyProtectedMember
    class _Decorator:
        @staticmethod
        def homotypic_binary_operator(inferior: typing.Callable[['Set', 'Set'], OperatorOutput]) \
                -> typing.Callable[['Set', 'Set'], OperatorOutput]:
            def wrapper(self: 'Set', other: 'Set') -> OperatorOutput:
                assert isinstance(self, Set) and isinstance(other, Set)
                if self.element_type == other.element_type:
                    return inferior(self, other)
                else:
                    raise InvalidOperandError('The requested binary operator is defined only for sets '
                                              'that share the same element type. The different types are: %r, %r' %
                                              (self.element_type.TYPE_NAME, other.element_type.TYPE_NAME))
            return wrapper

    def __init__(self, elements: typing.Iterable[Any]):
        list_of_elements = list(elements)   # type: typing.List[Any]
        del elements
        if len(list_of_elements) < 1:
            raise InvalidOperandError('Zero-length sets are currently not permitted because '
                                      'of associated type deduction issues. This may change later.')

        element_types = set(map(type, list_of_elements))
        if len(element_types) != 1:
            # This also weeds out covariant sets, although our barbie-size type system is unaware of that.
            raise InvalidOperandError('Heterogeneous sets are not permitted')

        self._element_type = list(element_types)[0]  # type: typing.Type[Any]
        self._value = frozenset(list_of_elements)    # type: typing.FrozenSet[Any]

        if not issubclass(self._element_type, Any):
            raise ValueError('Invalid element type: %r' % self._element_type)

    def __iter__(self) -> typing.Iterator[typing.Any]:
        return iter(self._value)

    @property
    def element_type(self) -> typing.Type[Any]:
        return self._element_type

    def __hash__(self) -> int:
        return hash(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Set):
            return self._value == other._value
        else:
            return NotImplemented

    def __str__(self) -> str:
        return '{%s}' % ', '.join(map(str, self._value))    # This is recursive.

    #
    # Set algebra implementation.
    #
    @_Decorator.homotypic_binary_operator
    def _is_equal_to(self, right: 'Set') -> bool:
        return self._value == right._value

    @_Decorator.homotypic_binary_operator
    def _is_superset_of(self, right: 'Set') -> bool:
        return self._value.issuperset(right._value)

    @_Decorator.homotypic_binary_operator
    def _is_subset_of(self, right: 'Set') -> bool:
        return self._value.issubset(right._value)

    @_Decorator.homotypic_binary_operator
    def _is_proper_superset_of(self, right: 'Set') -> bool:
        return self._is_superset_of(right) and not self._is_equal_to(right)

    @_Decorator.homotypic_binary_operator
    def _is_proper_subset_of(self, right: 'Set') -> bool:
        return self._is_subset_of(right) and not self._is_equal_to(right)

    @_Decorator.homotypic_binary_operator
    def _create_union_with(self, right: 'Set') -> 'Set':
        return Set(self._value.union(right._value))

    @_Decorator.homotypic_binary_operator
    def _create_intersection_with(self, right: 'Set') -> 'Set':
        return Set(self._value.intersection(right._value))

    @_Decorator.homotypic_binary_operator
    def _create_disjunctive_union_with(self, right: 'Set') -> 'Set':
        return Set(self._value.symmetric_difference(right._value))

    #
    # Set comparison.
    #
    def _equal(self, right: 'Any') -> 'Boolean':
        if isinstance(right, Set):
            return Boolean(self._is_equal_to(right))
        else:
            raise UndefinedOperatorError

    def _less_or_equal(self, right: 'Any') -> 'Boolean':
        if isinstance(right, Set):
            return Boolean(self._is_subset_of(right))
        else:
            raise UndefinedOperatorError

    def _greater_or_equal(self, right: 'Any') -> 'Boolean':
        if isinstance(right, Set):
            return Boolean(self._is_superset_of(right))
        else:
            raise UndefinedOperatorError

    def _less(self, right: 'Any') -> 'Boolean':
        if isinstance(right, Set):
            return Boolean(self._is_proper_subset_of(right))
        else:
            raise UndefinedOperatorError

    def _greater(self, right: 'Any') -> 'Boolean':
        if isinstance(right, Set):
            return Boolean(self._is_proper_superset_of(right))
        else:
            raise UndefinedOperatorError

    #
    # Set algebra operators that yield a new set.
    #
    def _bitwise_or(self, right: 'Any') -> 'Set':
        if isinstance(right, Set):
            return self._create_union_with(right)
        else:
            raise UndefinedOperatorError

    def _bitwise_xor(self, right: 'Any') -> 'Set':
        if isinstance(right, Set):
            return self._create_disjunctive_union_with(right)
        else:
            raise UndefinedOperatorError

    def _bitwise_and(self, right: 'Any') -> 'Set':
        if isinstance(right, Set):
            return self._create_intersection_with(right)
        else:
            raise UndefinedOperatorError

    #
    # Elementwise application.
    # https://stackoverflow.com/questions/55148139/referring-to-a-pure-virtual-method
    #
    def _elementwise(self,
                     impl: typing.Callable[['Any', 'Any'], 'Any'],
                     other: 'Any',
                     swap: bool = False) -> 'Set':
        if not isinstance(other, Set):
            return Set((impl(other, x) if swap else impl(x, other)) for x in self)
        else:
            raise UndefinedOperatorError

    def _add(self, right: 'Any') -> 'Set':
        return self._elementwise(add, right)

    def _add_right(self, left: 'Any') -> 'Set':
        return self._elementwise(add, left, swap=True)

    def _subtract(self, right: 'Any') -> 'Set':
        return self._elementwise(subtract, right)

    def _subtract_right(self, left: 'Any') -> 'Set':
        return self._elementwise(subtract, left, swap=True)

    def _multiply(self, right: 'Any') -> 'Set':
        return self._elementwise(multiply, right)

    def _multiply_right(self, left: 'Any') -> 'Set':
        return self._elementwise(multiply, left, swap=True)

    def _divide(self, right: 'Any') -> 'Set':
        return self._elementwise(divide, right)

    def _divide_right(self, left: 'Any') -> 'Set':
        return self._elementwise(divide, left, swap=True)

    def _modulo(self, right: 'Any') -> 'Set':
        return self._elementwise(modulo, right)

    def _modulo_right(self, left: 'Any') -> 'Set':
        return self._elementwise(modulo, left, swap=True)

    def _power(self, right: 'Any') -> 'Set':
        return self._elementwise(power, right)

    def _power_right(self, left: 'Any') -> 'Set':
        return self._elementwise(power, left, swap=True)

    #
    # Attributes
    #
    def _attribute(self, name: 'String') -> 'Any':
        if name.native_value == 'min':
            out = functools.reduce(lambda a, b: a if less(a, b) else b, self)
            assert isinstance(out, self.element_type)
        elif name.native_value == 'max':
            out = functools.reduce(lambda a, b: a if greater(a, b) else b, self)
            assert isinstance(out, self.element_type)
        elif name.native_value == 'count':  # "size" and "length" can be ambiguous, "cardinality" is long
            out = Rational(len(self._value))
        else:
            out = super(Set, self)._attribute(name)  # Hand over up the inheritance chain, this is important

        assert isinstance(out, Any)
        return out


#
# Operator wrappers. These wrappers serve two purposes:
#   - Late binding, as explained here: https://stackoverflow.com/questions/55148139/referring-to-a-pure-virtual-method
#   - Automatic left-right operand swapping when necessary (for some polyadic operators).
#
def _auto_swap(alternative_operator_name: typing.Optional[str] = None) -> \
        typing.Callable[[BinaryOperator[OperatorOutput]], BinaryOperator[OperatorOutput]]:
    def decorator(direct_operator: BinaryOperator[OperatorOutput]) -> BinaryOperator[OperatorOutput]:
        if alternative_operator_name:
            alternative_method_name = '_' + alternative_operator_name
        else:
            alternative_method_name = '_%s_right' % direct_operator.__name__

        if not hasattr(Any, alternative_method_name):  # pragma: no cover
            raise TypeError('The following alternative operator method is not defined: %r' % alternative_method_name)

        @functools.wraps(direct_operator)
        def wrapper(left: Any, right: Any) -> Any:
            if not isinstance(left, Any) or not isinstance(right, Any):  # pragma: no cover
                raise ValueError('Operators are only defined for implementations of Any; found this: %r, %r' %
                                 (type(left).__name__, type(right).__name__))
            try:
                result = direct_operator(left, right)
            except UndefinedOperatorError:
                if type(left) != type(right):
                    result = getattr(right, alternative_method_name)(left)  # Left and Right are swapped.
                else:
                    raise

            assert isinstance(result, Any)
            return result
        return wrapper
    return decorator


def logical_not(operand: Any) -> Boolean:                       # noinspection PyProtectedMember
    return operand._logical_not()


def positive(operand: Any) -> Any:                              # noinspection PyProtectedMember
    return operand._positive()


def negative(operand: Any) -> Any:                              # noinspection PyProtectedMember
    return operand._negative()


@_auto_swap('logical_or')  # Commutative
def logical_or(left: Any, right: Any) -> Boolean:               # noinspection PyProtectedMember
    return left._logical_or(right)


@_auto_swap('logical_and')  # Commutative
def logical_and(left: Any, right: Any) -> Boolean:              # noinspection PyProtectedMember
    return left._logical_and(right)


@_auto_swap('equal')  # Commutative
def equal(left: Any, right: Any) -> Boolean:                    # noinspection PyProtectedMember
    return left._equal(right)


# Special case - synthetic operator.
def not_equal(left: Any, right: Any) -> Boolean:                # noinspection PyProtectedMember
    return logical_not(equal(left, right))


@_auto_swap('greater_or_equal')
def less_or_equal(left: Any, right: Any) -> Boolean:            # noinspection PyProtectedMember
    return left._less_or_equal(right)


@_auto_swap('less_or_equal')
def greater_or_equal(left: Any, right: Any) -> Boolean:         # noinspection PyProtectedMember
    return left._greater_or_equal(right)


@_auto_swap('greater')
def less(left: Any, right: Any) -> Boolean:                     # noinspection PyProtectedMember
    return left._less(right)


@_auto_swap('less')
def greater(left: Any, right: Any) -> Boolean:                  # noinspection PyProtectedMember
    return left._greater(right)


@_auto_swap()
def bitwise_or(left: Any, right: Any) -> Any:                   # noinspection PyProtectedMember
    return left._bitwise_or(right)


@_auto_swap()
def bitwise_xor(left: Any, right: Any) -> Any:                  # noinspection PyProtectedMember
    return left._bitwise_xor(right)


@_auto_swap()
def bitwise_and(left: Any, right: Any) -> Any:                  # noinspection PyProtectedMember
    return left._bitwise_and(right)


@_auto_swap()
def add(left: Any, right: Any) -> Any:                          # noinspection PyProtectedMember
    return left._add(right)


@_auto_swap()
def subtract(left: Any, right: Any) -> Any:                     # noinspection PyProtectedMember
    return left._subtract(right)


@_auto_swap()
def multiply(left: Any, right: Any) -> Any:                     # noinspection PyProtectedMember
    return left._multiply(right)


@_auto_swap()
def divide(left: Any, right: Any) -> Any:                       # noinspection PyProtectedMember
    return left._divide(right)


@_auto_swap()
def modulo(left: Any, right: Any) -> Any:                       # noinspection PyProtectedMember
    return left._modulo(right)


@_auto_swap()
def power(left: Any, right: Any) -> Any:                        # noinspection PyProtectedMember
    return left._power(right)


# Special case - no argument-swapped alternative defined.
# We accept both native strings and String in order to support both dynamically computed attributes and
# statically defined attributes.
def attribute(value: Any, name: typing.Union[str, String]) -> Any:
    if isinstance(name, str):
        name = String(name)

    if isinstance(value, Any) and isinstance(name, String):     # noinspection PyProtectedMember
        return value._attribute(name)
    else:  # pragma: no cover
        raise ValueError('The argument types of the attribute operator are (Any, String), got (%r, %r)' %
                         (type(value).__name__, type(name).__name__))


# noinspection PyUnresolvedReferences,PyTypeChecker
def _unittest_expressions() -> None:
    r = Rational
    s = String

    for a in (True, False):
        for b in (True, False):
            assert Boolean(a).native_value == a
            assert logical_not(Boolean(a)).native_value == (not a)
            assert logical_and(Boolean(a), Boolean(b)).native_value == (a and b)
            assert logical_or(Boolean(a), Boolean(b)).native_value == (a or b)

    assert \
        equal(
            divide(
                multiply(
                    add(r(2), r(2)),
                    r(3)
                ),
                r(5)
            ),
            r(fractions.Fraction(12, 5))
        ).native_value

    assert add(s('123'), s('abc')).native_value == '123abc'  # type: ignore

    new_set = add(Set([s('123'), s('456')]),
                  s('abc'))
    assert set(new_set) == {s('123abc'), s('456abc')}  # type: ignore

    new_set = add(s('abc'),
                  Set([s('123'), s('456')]))
    assert set(new_set) == {s('abc123'), s('abc456')}  # type: ignore

    new_set = add(s('abc'),
                  Set([Set([s('123'), s('456')]),
                       Set([s('789'), s('987')])]))
    assert new_set == Set([Set([s('abc123'), s('abc456')]),
                           Set([s('abc789'), s('abc987')])])

    assert attribute(Set([r(1), r(2), r(3), r(-4), r(-5)]), s('min')) == r(-5)
    assert attribute(Set([r(1), r(2), r(3), r(-4), r(-5)]), s('max')) == r(3)


def _unittest_textual_representations() -> None:
    assert str(Rational(fractions.Fraction(123, 456))) == '41/152'
    assert repr(Rational(fractions.Fraction(123, 456))) == 'rational(41/152)'
    assert str(Rational(-123)) == '-123'
    assert repr(Rational(-123)) == 'rational(-123)'

    assert str(Boolean(True)) == 'true'
    assert repr(Boolean(False)) == 'bool(false)'

    assert str(String('Hello\nworld!')) == r"'Hello\nworld!'"
    assert repr(String('Hello\nworld!')) == r"string('Hello\nworld!')"

    tmp = str(Set([Rational(1), Rational(fractions.Fraction(-9, 7))]))
    assert tmp == '{1, -9/7}' or tmp == '{-9/7, 1}'

    tmp = repr(Set([Rational(1), Rational(fractions.Fraction(-9, 7))]))
    assert tmp == 'set({1, -9/7})' or tmp == 'set({-9/7, 1})'

    tmp = str(Set([Set([Rational(1), Rational(fractions.Fraction(-9, 7))]),
                   Set([Rational(fractions.Fraction(90, 7))])]))
    assert \
        tmp == '{{1, -9/7}, {90/7}}' or \
        tmp == '{{-9/7, 1}, {90/7}}' or \
        tmp == '{{90/7}, {-9/7, 1}}' or \
        tmp == '{{90/7}, {1, -9/7}}'

    assert repr(Set([String('123')])) == "set({'123'})"


# noinspection PyTypeChecker
def _unittest_basic() -> None:
    from pytest import raises

    assert hash(Boolean(True)) == hash(True)
    assert Boolean(True) == Boolean(True)
    assert Boolean(True) != Boolean(False)
    assert Boolean(True) != Rational(1)         # sic!
    assert Boolean(True) != Rational(123)
    assert Boolean(True) != Set([Boolean(True)])

    with raises(ValueError):
        Boolean(int)       # type: ignore

    with raises(ValueError):
        Rational({123})    # type: ignore

    with raises(ValueError):
        Rational('123')    # type: ignore

    with raises(ValueError):
        String(123)       # type: ignore

    with raises(ValueError):
        Set([123])        # type: ignore

    assert Rational(123).is_integer()
    assert not Rational(fractions.Fraction(123, 124)).is_integer()
    assert Rational(-123).as_native_integer() == -123
    with raises(InvalidOperandError):
        Rational(fractions.Fraction(123, 124)).as_native_integer()
